from docutils import nodes
from docutils.transforms import Transform


class LineNumberTransform(Transform):
    default_priority = 500

    def apply(self, **kwargs):
        for node in self.document.traverse(nodes.paragraph):
            if node.line:
                node["classes"].append("linemarker")
                node["classes"].append(f"linemarker-{node.line}")


def setup(app):
    app.add_transform(LineNumberTransform)



"""Utility functions to help with testing Language Server features."""
import logging
import pathlib
import re
from typing import List
from typing import Optional
from typing import Union

import pygls.uris as Uri
from lsprotocol.types import ClientCapabilities
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionList
from lsprotocol.types import CompletionParams
from lsprotocol.types import DidChangeTextDocumentParams
from lsprotocol.types import DidCloseTextDocumentParams
from lsprotocol.types import DidOpenTextDocumentParams
from lsprotocol.types import Hover
from lsprotocol.types import HoverParams
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import TextDocumentContentChangeEvent_Type1
from lsprotocol.types import TextDocumentIdentifier
from lsprotocol.types import TextDocumentItem
from lsprotocol.types import VersionedTextDocumentIdentifier
from pygls.workspace import Document
from pytest_lsp import LanguageClient
from pytest_lsp import make_test_lsp_client
from sphinx import __version__ as __sphinx_version__

from esbonio.lsp import CompletionContext
from esbonio.lsp.rst.config import ServerCompletionConfig

logger = logging.getLogger(__name__)


def _noop(*args, **kwargs):
    ...


def make_esbonio_client(*args, **kwargs) -> LanguageClient:
    """Construct a pytest-lsp client that is aware of esbonio specific messages"""
    client = make_test_lsp_client(*args, **kwargs)
    client.feature("esbonio/buildStart")(_noop)
    client.feature("esbonio/buildComplete")(_noop)

    return client


def sphinx_version(
    eq: Optional[int] = None,
    lt: Optional[int] = None,
    lte: Optional[int] = None,
    gt: Optional[int] = None,
    gte: Optional[int] = None,
) -> bool:
    """Helper function for determining which version of Sphinx we are
    testing with.

    .. note::

       Currently this function only considers the major version number.

    Parameters
    ----------
    eq
       When set, this function returns ``True`` if Sphinx's version is exactly
       what's given.

    gt
       When set, this function returns ``True`` if Sphinx's version is strictly
       greater than what's given

    gte
       When set, this function returns ``True`` if Sphinx's version is greater than
       or equal to what's given

    lt
       When set, this function returns ``True`` if Sphinx's version is strictly
       less than what's given

    lte
       When set, this function returns ``True`` if Sphinx's version is less than
       or equal to what's given

    """

    major, _, _ = (int(v) for v in __sphinx_version__.split("."))

    if eq is not None:
        return major == eq

    if gt is not None:
        return major > gt

    if gte is not None:
        return major >= gte

    if lt is not None:
        return major < lt

    if lte is not None:
        return major <= lte

    return False


def range_from_str(spec: str) -> Range:
    """Create a range from the given string ``a:b-x:y``"""
    start, end = spec.split("-")
    sl, sc = start.split(":")
    el, ec = end.split(":")

    return Range(
        start=Position(line=int(sl), character=int(sc)),
        end=Position(line=int(el), character=int(ec)),
    )


def make_completion_context(
    pattern: re.Pattern,
    text: str,
    *,
    character: int = -1,
    prefer_insert: bool = False,
) -> CompletionContext:
    """Helper for making test completion context instances.

    Parameters
    ----------
    pattern
       The regular expression pattern that corresponds to the completion request.

    text
       The text that "triggered" the completion request

    character
       The character column at which the request is being made.
       If ``-1`` (the default), it will be assumed that the request is being made at
       the end of ``text``.

    prefer_insert
       Flag to indicate if the ``preferred_insert_behavior`` option should be set to
       ``insert``
    """

    match = pattern.match(text)
    if not match:
        raise ValueError(f"'{text}' is not valid in this completion context")

    line = 0
    character = len(text) if character == -1 else character

    return CompletionContext(
        doc=Document(uri="file:///test.txt"),
        location="rst",
        match=match,
        position=Position(line=line, character=character),
        config=ServerCompletionConfig(
            preferred_insert_behavior="insert" if prefer_insert else "replace"
        ),
        capabilities=ClientCapabilities(),
    )


def directive_argument_patterns(name: str, partial: str = "") -> List[str]:
    """Return a number of example directive argument patterns.

    These correspond to test cases where directive argument suggestions should be
    generated.

    Parameters
    ----------
    name:
       The name of the directive to generate suggestions for.
    partial:
       The partial argument that the user has already entered.
    """
    return [s.format(name, partial) for s in [".. {}:: {}", "   .. {}:: {}"]]


def role_patterns(partial: str = "") -> List[str]:
    """Return a number of example role patterns.

    These correspond to when role suggestions should be generated.

    Parameters
    ----------
    partial:
       The partial role name that the user has already entered
    """
    return [
        s.format(partial)
        for s in [
            "{}",
            "({}",
            "- {}",
            "   {}",
            "   ({}",
            "   - {}",
            "some text {}",
            "some text ({}",
            "   some text {}",
            "   some text ({}",
        ]
    ]


def role_target_patterns(
    name: str, partial: str = "", include_modifiers: bool = True
) -> List[str]:
    """Return a number of example role target patterns.

    These correspond to test cases where role target suggestions should be generated.

    Parameters
    ----------
    name:
       The name of the role to generate suggestions for.
    partial:
       The partial target that the user as already entered.
    include_modifiers:
       A flag to indicate if additional modifiers like ``!`` and ``~`` should be
       included in the generated patterns.
    """

    patterns = [
        ":{}:`{}",
        "(:{}:`{}",
        "- :{}:`{}",
        ":{}:`More Info <{}",
        "(:{}:`More Info <{}",
        "   :{}:`{}",
        "   (:{}:`{}",
        "   - :{}:`{}",
        "   :{}:`Some Label <{}",
        "   (:{}:`Some Label <{}",
    ]

    test_cases = [p.format(name, partial) for p in patterns]

    if include_modifiers:
        test_cases += [p.format(name, "!" + partial) for p in patterns]
        test_cases += [p.format(name, "~" + partial) for p in patterns]

    return test_cases


def intersphinx_target_patterns(name: str, project: str) -> List[str]:
    """Return a number of example intersphinx target patterns.

    These correspond to cases where target completions may be generated

    Parameters
    ----------
    name: str
       The name of the role to generate examples for
    project: str
       The name of the project to generate examples for
    """
    return [
        s.format(name, project)
        for s in [
            ":{}:`{}:",
            "(:{}:`{}:",
            ":{}:`More Info <{}:",
            "(:{}:`More Info <{}:",
            "   :{}:`{}:",
            "   (:{}:`{}:",
            "   :{}:`Some Label <{}:",
            "   (:{}:`Some Label <{}:",
        ]
    ]


async def completion_request(
    client: LanguageClient, test_uri: str, text: str, character: Optional[int] = None
) -> Union[CompletionList, List[CompletionItem], None]:
    """Make a completion request to a language server.

    Intended for use within test cases, this function simulates the opening of a
    document, inserting some text, triggering a completion request and closing it
    again.

    The file referenced by ``test_uri`` does not have to exist.

    The text to be inserted is specified through the ``text`` parameter. By default
    it's assumed that the ``text`` parameter consists of a single line of text, in fact
    this function will error if that is not the case.

    If your request requires additional context (such as directive option completions)
    it can be included but it must be delimited with a ``\\f`` character. For example,
    to represent the following scenario::

       .. image:: filename.png
          :align: center
          :
           ^

    where ``^`` represents the position from which we trigger the completion request.
    We would set ``text`` to the following
    ``.. image:: filename.png\\n   :align: center\\n\\f   :``

    Parameters
    ----------
    test:
       The client used to make the request.
    test_uri:
       The uri the completion request should be made within.
    text
       The text that provides the context for the completion request.
    character:
       The character index at which to make the completion request from.
       If ``None``, it will default to the end of the inserted text.
    """

    if "\f" in text:
        contents, text = text.split("\f")
    else:
        contents = ""

    logger.debug("Context text:    '%s'", contents)
    logger.debug("Insertion text: '%s'", text)
    assert "\n" not in text, "Insertion text cannot contain newlines"

    ext = pathlib.Path(Uri.to_fs_path(test_uri)).suffix
    lang_id = "python" if ext == ".py" else "rst"

    client.text_document_did_open(
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri, language_id=lang_id, version=1, text=contents
            )
        )
    )

    lines = contents.split("\n")
    line = len(lines) - 1
    insertion_point = len(lines[-1])

    new_lines = text.split("\n")
    num_new_lines = len(new_lines) - 1
    num_new_chars = len(new_lines[-1])

    if num_new_lines > 0:
        end_char = num_new_chars
    else:
        end_char = insertion_point + num_new_chars

    client.text_document_did_change(
        DidChangeTextDocumentParams(
            text_document=VersionedTextDocumentIdentifier(uri=test_uri, version=2),
            content_changes=[
                TextDocumentContentChangeEvent_Type1(
                    text=text,
                    range=Range(
                        start=Position(line=line, character=insertion_point),
                        end=Position(line=line + num_new_lines, character=end_char),
                    ),
                )
            ],
        )
    )

    character = character or insertion_point + len(text)
    response = await client.text_document_completion_async(
        CompletionParams(
            text_document=TextDocumentIdentifier(uri=test_uri),
            position=Position(line=line, character=character),
        )
    )

    client.text_document_did_close(
        DidCloseTextDocumentParams(text_document=TextDocumentIdentifier(uri=test_uri))
    )

    return response


async def hover_request(
    client: LanguageClient, test_uri: str, text: str, line: int, character: int
) -> Optional[Hover]:
    """Make a hover request to a language server.

    Intended for use within test cases, this function simulates the opening of a
    document containing some text, triggering a hover request and closing it again.

    The file referenced by ``test_uri`` does not have to exist.

    Parameters
    ----------
    test
       The client used to make the request.

    test_uri
       The uri the completion request should be made within.

    text
       The text that provides the context for the hover request.

    line
       The line number to make the hover request from

    character
       The column number to make the hover request from
    """
    ext = pathlib.Path(Uri.to_fs_path(test_uri)).suffix
    lang_id = "python" if ext == ".py" else "rst"

    client.text_document_did_open(
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri, language_id=lang_id, version=1, text=text
            )
        )
    )

    response = await client.text_document_hover_async(
        HoverParams(
            text_document=TextDocumentIdentifier(uri=test_uri),
            position=Position(line=line, character=character),
        )
    )

    client.text_document_did_close(
        DidCloseTextDocumentParams(text_document=TextDocumentIdentifier(uri=test_uri))
    )

    return response



"""Role support."""
import typing
import warnings
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from lsprotocol.types import CompletionItem
from lsprotocol.types import DocumentLink
from lsprotocol.types import Location
from lsprotocol.types import MarkupContent
from lsprotocol.types import MarkupKind
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import TextEdit

from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import DefinitionContext
from esbonio.lsp.rst import DocumentLinkContext
from esbonio.lsp.rst import HoverContext
from esbonio.lsp.rst import ImplementationContext
from esbonio.lsp.rst import LanguageFeature
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.inspect import get_object_location
from esbonio.lsp.util.patterns import DEFAULT_ROLE
from esbonio.lsp.util.patterns import DIRECTIVE
from esbonio.lsp.util.patterns import ROLE

from .completions import render_role_completion

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore[assignment]


class RoleLanguageFeature:
    """Base class for role language features."""

    def complete_targets(
        self, context: CompletionContext, name: str, domain: str
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        role.

        Parameters
        ----------
        context
           The completion context

        name
           The name of the role to generate completion suggestions for.

        domain
           The name of the domain the role is a member of
        """
        return []

    def find_target_definitions(
        self, context: DefinitionContext, name: str, domain: str, label: str
    ) -> List[Location]:
        """Return a list of locations representing the definition of the given role
        target.

        Parameters
        ----------
        doc:
           The document containing the match
        match:
           The match object that triggered the definition request
        name:
           The name of the role
        domain:
           The domain the role is part of, if applicable.
        """
        return []

    def get_implementation(self, role: str, domain: str) -> Optional[Any]:
        """Return the implementation for the given role name.

        Parameters
        ----------
        role
           The name of the role

        domain
           The domain the role belongs to, if any
        """
        return self.index_roles().get(role, None)

    def index_roles(self) -> Dict[str, Any]:
        """Return all known roles."""
        return dict()

    def resolve_target_link(
        self, context: DocumentLinkContext, name: str, domain: Optional[str], label: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return a link corresponding to the given target.

        Parameters
        ----------
        context
           The document link context

        domain
           The name (if applicable) of the domain the role is a member of

        name
           The name of the role to generate completion suggestions for.

        label
           The label of the target to provide the link for
        """
        return None, None

    def suggest_roles(self, context: CompletionContext) -> Iterable[Tuple[str, Any]]:
        """Suggest roles that may be used, given a completion context."""
        return self.index_roles().items()


class TargetDefinition(Protocol):
    """A definition provider for role targets.

    .. deprecated:: 0.15.0

       This will be removed in ``v1.0``, use a subclass of
       :class:`~esbonio.lsp.roles.RoleLanguageFeature` instead.
    """

    def find_definitions(
        self, context: DefinitionContext, name: str, domain: Optional[str]
    ) -> List[Location]:
        """Return a list of locations representing the definition of the given role
        target.

        Parameters
        ----------
        doc:
           The document containing the match
        match:
           The match object that triggered the definition request
        name:
           The name of the role
        domain:
           The domain the role is part of, if applicable.
        """


class TargetCompletion(Protocol):
    """A completion provider for role targets.

    .. deprecated:: 0.15.0

       This will be removed in ``v1.0``, use a subclass of
       :class:`~esbonio.lsp.roles.RoleLanguageFeature` instead.
    """

    def complete_targets(
        self, context: CompletionContext, name: str, domain: Optional[str]
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        role.

        Parameters
        ----------
        context:
           The completion context
        domain:
           The name of the domain the role is a member of
        name:
           The name of the role to generate completion suggestions for.
        """


class TargetLink(Protocol):
    """A document link provider for role targets.

    .. deprecated:: 0.15.0

       This will be removed in ``v1.0``, use a subclass of
       :class:`~esbonio.lsp.roles.RoleLanguageFeature` instead.
    """

    def resolve_link(
        self, context: DocumentLinkContext, name: str, domain: Optional[str], label: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Return a link corresponding to the given target.

        Parameters
        ----------
        context
           The document link context

        domain
           The name (if applicable) of the domain the role is a member of

        name
           The name of the role to generate completion suggestions for.

        label
           The label of the target to provide the link for
        """


class Roles(LanguageFeature):
    """Role support for the language server."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._documentation: Dict[str, Dict[str, str]] = {}
        """Cache for documentation."""

        self._features: Dict[str, RoleLanguageFeature] = {}
        """Collection of registered features"""

    def add_feature(self, feature: RoleLanguageFeature):
        """Register a role language feature

        Parameters
        ----------
        feature
           The role language feature
        """
        key = f"{feature.__module__}.{feature.__class__.__name__}"

        # Create a unique key for this instance.
        if key in self._features:
            key += f".{len([k for k in self._features.keys() if k.startswith(key)])}"

        self._features[key] = feature

    def add_target_definition_provider(self, provider: TargetDefinition) -> None:
        """Register a :class:`~esbonio.lsp.roles.TargetDefinition` provider.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0`` use
           :meth:`~esbonio.lsp.roles.Roles.add_feature` with a
           :class:`~esbonio.lsp.roles.RoleLanguageFeature` subclass instead.

        Parameters
        ----------
        provider
           The provider to register
        """

        warnings.warn(
            "TargetDefinition providers are deprecated in favour of "
            "RoleLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.definition"

        def find_target_definitions(self, context, name, domain, label):
            return provider.find_definitions(context, name, domain)

        feature = type(
            f"{name}TargetDefinitionProvider",
            (RoleLanguageFeature,),
            {"find_target_definitions": find_target_definitions},
        )()

        self._features[key] = feature

    def add_target_link_provider(self, provider: TargetLink) -> None:
        """Register a :class:`~esbonio.lsp.roles.TargetLink` provider.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0`` use
           :meth:`~esbonio.lsp.roles.Roles.add_feature` with a
           :class:`~esbonio.lsp.roles.RoleLanguageFeature` subclass instead.

        Parameters
        ----------
        provider
           The provider to register
        """

        warnings.warn(
            "TargetLink providers are deprecated in favour of "
            "RoleLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.link"

        feature = type(
            f"{name}TargetLinkProvider",
            (RoleLanguageFeature,),
            {"resolve_target_link": provider.resolve_link},
        )()

        self._features[key] = feature

    def add_target_completion_provider(self, provider: TargetCompletion) -> None:
        """Register a :class:`~esbonio.lsp.roles.TargetCompletion` provider.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0`` use
           :meth:`~esbonio.lsp.roles.Roles.add_feature` with a
           :class:`~esbonio.lsp.roles.RoleLanguageFeature` subclass instead.

        Parameters
        ----------
        provider
           The provider to register
        """

        warnings.warn(
            "TargetCompletion providers are deprecated in favour of "
            "RoleLanguageFeatures, this method will be removed in v1.0",
            DeprecationWarning,
            stacklevel=2,
        )

        name = provider.__class__.__name__
        key = f"{provider.__module__}.{name}.completion"

        feature = type(
            f"{name}TargetCompletionProvider",
            (RoleLanguageFeature,),
            {"complete_targets": provider.complete_targets},
        )()

        self._features[key] = feature

    def add_documentation(self, documentation: Dict[str, Dict[str, Any]]) -> None:
        """Register role documentation.

        ``documentation`` should be a dictionary of the form ::

           documentation = {
               "raw(docutils.parsers.rst.roles.raw_role)": {
                   "is_markdown": true,
                   "license": "https://...",
                   "source": "https://...",
                   "description": [
                       "# :raw:",
                       "The raw role is used for...",
                       ...
                   ]
               }
           }

        where the key is of the form `name(dotted_name)`. There are cases where a role's
        implementation is not sufficient to uniquely identify it as multiple roles can
        be provided by a single class.

        This means the key has to be a combination of the ``name`` the user writes in
        an reStructuredText document and ``dotted_name`` is the fully qualified name of
        the role's implementation.

        .. note::

           If there is a clash with an existing key, the existing value will be
           overwritten with the new value.

        The values in this dictionary are themselves dictionaries with the following
        fields.

        ``description``
           A list of strings for the role's usage.

        ``is_markdown``
           A boolean flag used to indicate whether the ``description`` is written in
           plain text or markdown.

        ``source``
           The url to the documentation's source.

        ``license``
           The url to the documentation's license.

        Parameters
        ----------
        documentation:
           The documentation to register.
        """

        for key, doc in documentation.items():
            description = doc.get("description", [])
            if not description:
                continue

            source = doc.get("source", "")
            if source:
                description.append(f"\n[Source]({source})")

            license = doc.get("license", "")
            if license:
                description.append(f"\n[License]({license})")

            doc["description"] = "\n".join(description)
            self._documentation[key] = doc

    completion_triggers = [ROLE, DEFAULT_ROLE]
    definition_triggers = [ROLE]
    hover_triggers = [ROLE]
    implementation_triggers = [ROLE]

    def definition(self, context: DefinitionContext) -> List[Location]:
        domain = context.match.group("domain") or ""
        name = context.match.group("name")
        label = context.match.group("label")

        # Be sure to only match complete roles
        if not label or not context.match.group(0).endswith("`"):
            return []

        return self.find_target_definitions(context, name, domain, label)

    def find_target_definitions(
        self, context: DefinitionContext, name: str, domain: str, label: str
    ) -> List[Location]:
        definitions = []

        for feature_name, feature in self._features.items():
            try:
                definitions += feature.find_target_definitions(
                    context, name, domain, label
                )
            except Exception:
                self.logger.error(
                    "Unable to find definitions of '%s' for role ':%s:', "
                    "error in feature: '%s'",
                    label,
                    f"{domain}:{name}" if domain else name,
                    feature_name,
                    exc_info=True,
                )

        return definitions

    def document_link(self, context: DocumentLinkContext) -> List[DocumentLink]:
        links = []

        for line, text in enumerate(context.doc.lines):
            for match in ROLE.finditer(text):
                label = match.group("label")

                # Be sure to only match complete roles
                if not label or not match.group(0).endswith("`"):
                    continue

                domain = match.group("domain")
                name = match.group("name")

                target, tooltip = self.resolve_target_link(context, name, domain, label)
                if not target:
                    continue

                idx = match.group(0).index(label)
                start = match.start() + idx
                end = start + len(label)

                link = DocumentLink(
                    target=target,
                    tooltip=tooltip if context.tooltip_support else None,
                    range=Range(
                        start=Position(line=line, character=start),
                        end=Position(line=line, character=end),
                    ),
                )

                links.append(link)

        return links

    def resolve_target_link(
        self, context: DocumentLinkContext, name: str, domain: str, label: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Resolve a given document link."""

        for feature_name, feature in self._features.items():
            try:
                target, tooltip = feature.resolve_target_link(
                    context, name, domain, label
                )

                if target:
                    return target, tooltip
            except Exception:
                self.logger.error(
                    "Unable to resolve target link '%s' for role ':%s:', "
                    "error in feature: '%s'",
                    label,
                    f"{domain}:{name}" if domain else name,
                    feature_name,
                    exc_info=True,
                )

        return None, None

    def complete(self, context: CompletionContext) -> List[CompletionItem]:
        """Generate completion suggestions relevant to the current context.

        This function is a little intense, but its sole purpose is to determine the
        context in which the completion request is being made and either return
        nothing, or the results of :meth:`~esbonio.lsp.roles.Roles.complete_roles` or
        :meth:`esbonio.lsp.roles.Roles.complete_targets` whichever is appropriate.

        Parameters
        ----------
        context:
           The context of the completion request.
        """

        # Do not suggest completions within the middle of Python code.
        if context.location == "py":
            return []

        groups = context.match.groupdict()
        target = groups["target"]

        # All text matched by the regex
        text = context.match.group(0)
        start, end = context.match.span()

        if target:
            target_index = start + text.find(target)

            # Only trigger target completions if the request was made from within
            # the target part of the role.
            if target_index <= context.position.character <= end:
                return self.complete_targets(context)

        # If there's no indent, then this can only be a role definition
        indent = context.match.group(1)
        if indent == "":
            return self.complete_roles(context)

        # Otherwise, search backwards until we find a blank line or an unindent
        # so that we can determine the appropriate context.
        linum = context.position.line - 1

        try:
            line = context.doc.lines[linum]
        except IndexError:
            return self.complete_roles(context)

        while linum >= 0 and line.startswith(indent):
            linum -= 1
            line = context.doc.lines[linum]

        # Unless we are within a directive's options block, we should offer role
        # suggestions
        if DIRECTIVE.match(line):
            return []

        return self.complete_roles(context)

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:
        # We need extra info to know who to call
        if not item.data:
            return item

        data = typing.cast(Dict, item.data)
        ctype = data.get("completion_type", "")

        if ctype == "role":
            return self.completion_resolve_role(item)

        return item

    def suggest_roles(self, context: CompletionContext) -> Iterable[Tuple[str, Any]]:
        """Suggest roles that may be used, given a completion context.

        Parameters
        ----------
        context
           The completion context
        """
        for name, feature in self._features.items():
            try:
                yield from feature.suggest_roles(context)
            except Exception:
                self.logger.error(
                    "Unable to suggest roles, error in feature: '%s'",
                    name,
                    exc_info=True,
                )

    def complete_roles(self, context: CompletionContext) -> List[CompletionItem]:
        items = []

        for name, role in self.suggest_roles(context):
            item = render_role_completion(context, name, role)
            if item is None:
                continue

            items.append(item)

        return items

    def completion_resolve_role(self, item: CompletionItem) -> CompletionItem:
        # We need the detail field set to the role implementation's fully qualified name
        if not item.detail:
            return item

        documentation = self.get_documentation(item.label, item.detail)
        if not documentation:
            return item

        description = documentation.get("description", "")
        is_markdown = documentation.get("is_markdown", False)
        kind = MarkupKind.Markdown if is_markdown else MarkupKind.PlainText

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    def suggest_targets(
        self, context: CompletionContext, name: str, domain: str
    ) -> List[CompletionItem]:
        targets = []

        for feature_name, feature in self._features.items():
            try:
                targets += feature.complete_targets(context, name, domain)
            except Exception:
                self.logger.error(
                    "Unable to suggest targets for role ':%s:', error in feature: '%s'",
                    f"{domain}:{name}" if domain else name,
                    feature_name,
                    exc_info=True,
                )

        return targets

    def complete_targets(self, context: CompletionContext) -> List[CompletionItem]:
        """Generate the list of role target completion suggestions."""

        groups = context.match.groupdict()

        # Handle the default role case.
        if "role" not in groups:
            domain, name = self.rst.get_default_role()
            if not name:
                return []
        else:
            name = groups["name"]
            domain = groups["domain"]

        domain = domain or ""
        name = name or ""

        # Only generate suggestions for "aliased" targets if the request comes from
        # within the <> chars.
        if groups["alias"]:
            text = context.match.group(0)
            start = context.match.span()[0] + text.find(groups["alias"])
            end = start + len(groups["alias"])

            if start <= context.position.character <= end:
                return []

        targets = []

        startchar = "<" if "<" in groups["target"] else "`"
        endchars = ">`" if "<" in groups["target"] else "`"

        start, end = context.match.span()
        start += context.match.group(0).index(startchar) + 1
        range_ = Range(
            start=Position(line=context.position.line, character=start),
            end=Position(line=context.position.line, character=end),
        )
        prefix = context.match.group(0)[start:]
        modifier = groups["modifier"] or ""

        for candidate in self.suggest_targets(context, name, domain):
            # Don't interfere with items that already carry a `text_edit`, allowing
            # some providers (like filepaths) to do something special.
            if not candidate.text_edit:
                new_text = candidate.insert_text or candidate.label

                # This is rather annoying, but `filter_text` needs to start with
                # the text we are going to replace, otherwise VSCode won't show our
                # suggestions!
                candidate.filter_text = f"{prefix}{new_text}"

                candidate.text_edit = TextEdit(
                    range=range_, new_text=f"{modifier}{new_text}"
                )
                candidate.insert_text = None

            if not candidate.text_edit.new_text.endswith(endchars):
                candidate.text_edit.new_text += endchars

            targets.append(candidate)

        return targets

    def hover(self, context: HoverContext) -> str:
        if context.location not in {"rst", "docstring"}:
            return ""

        name = context.match.group("name")
        domain = context.match.group("domain")

        # Determine if the hover is on the :role: itself, or within the `target`.
        idx = context.position.character - context.match.start()
        prefix = context.match.group(0)[:idx]

        if "`" in prefix:
            return self.hover_target(context, name, domain)

        return self.hover_role(context, name, domain)

    def hover_role(self, context: HoverContext, name: str, domain: str) -> str:
        label = f"{domain}:{name}" if domain else name
        role = self.get_implementation(name, domain)
        if not role:
            return ""

        try:
            dotted_name = f"{role.__module__}.{role.__name__}"
        except AttributeError:
            dotted_name = f"{role.__module__}.{role.__class__.__name__}"

        documentation = self.get_documentation(label, dotted_name)
        if not documentation:
            return ""

        return documentation.get("description", "")

    def hover_target(
        self, context: HoverContext, name: str, domain: Optional[str]
    ) -> str:
        # TODO: Add extension point for providers to contribute hovers for a target.
        return ""

    def get_roles(self) -> Dict[str, Any]:
        """Return a dictionary of all known roles."""

        roles = {}

        for name, feature in self._features.items():
            self.logger.debug("calling '%s'", name)
            try:
                roles.update(feature.index_roles())
            except Exception:
                self.logger.error(
                    "Unable to index roles, error in feature '%s'", name, exc_info=True
                )

        return roles

    def get_implementation(self, role: str, domain: str) -> Optional[Any]:
        """Return the implementation of a role given its name

        Parameters
        ----------
        role
           The name of the role.

        domain
           The domain of the role, if applicable.
        """

        if domain:
            name = f"{domain}:{role}"
        else:
            name = role

        for feature_name, feature in self._features.items():
            try:
                impl = feature.get_implementation(role, domain)
                if impl is not None:
                    return impl
            except Exception:
                self.logger.error(
                    "Unable to get implementation for ':%s:', error in feature: '%s'\n%s",
                    name,
                    feature_name,
                    exc_info=True,
                )

        self.logger.debug("Unable to get implementation for ':%s:', unknown role", name)
        return None

    def implementation(self, context: ImplementationContext) -> List[Location]:
        region = context.match.group("role")
        name = context.match.group("name")
        domain = context.match.group("domain")

        start = context.match.start() + context.match.group(0).index(region)
        end = start + len(region)

        self.logger.debug("%s, %s, %s", region, name, domain)
        self.logger.debug("%s, %s", start, end)

        if start <= context.position.character <= end:
            return self.find_role_implementation(context, name, domain)

        return []

    def find_role_implementation(
        self, context: ImplementationContext, name: str, domain: str
    ) -> List[Location]:
        impl = self.get_implementation(name, domain)
        if impl is None:
            return []

        location = get_object_location(impl, self.logger)
        if location is not None:
            return [location]

        # Some roles are implemented as instances of some class
        location = get_object_location(impl.__class__, self.logger)
        if location is not None:
            return [location]

        return []

    def get_documentation(
        self, label: str, implementation: str
    ) -> Optional[Dict[str, Any]]:
        """Return the documentation for the given role, if available.

        If documentation for the given ``label`` cannot be found, this function will also
        look for the label under the project's :confval:`sphinx:primary_domain` followed
        by the ``std`` domain.

        Parameters
        ----------
        label:
           The name of the role, as the user would type in an reStructuredText file.
        implementation:
           The full dotted name of the role's implementation.
        """

        key = f"{label}({implementation})"
        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        if not isinstance(self.rst, SphinxLanguageServer) or not self.rst.app:
            return None

        # Nothing found, try the primary domain
        domain = self.rst.app.config.primary_domain
        key = f"{domain}:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        # Still nothing, try the standard domain
        key = f"std:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        return None


def esbonio_setup(rst: RstLanguageServer):
    rst.add_feature(Roles(rst))



import collections
import inspect
import logging
import pathlib
import re
import traceback
import typing
import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar

import pygls.uris as Uri
from docutils.parsers.rst import Directive
from lsprotocol.converters import get_converter
from lsprotocol.types import ClientCapabilities
from lsprotocol.types import CodeAction
from lsprotocol.types import CodeActionParams
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionItemTag
from lsprotocol.types import DeleteFilesParams
from lsprotocol.types import Diagnostic
from lsprotocol.types import DidSaveTextDocumentParams
from lsprotocol.types import DocumentLink
from lsprotocol.types import InitializedParams
from lsprotocol.types import InitializeParams
from lsprotocol.types import Location
from lsprotocol.types import MarkupKind
from lsprotocol.types import Position
from pygls import IS_WIN
from pygls.capabilities import get_capability
from pygls.server import LanguageServer
from pygls.workspace import Document

from esbonio.cli import setup_cli
from esbonio.lsp.log import setup_logging

from .config import InitializationOptions
from .config import ServerCompletionConfig
from .io import read_initial_doctree

converter = get_converter()
LF = TypeVar("LF", bound="LanguageFeature")
TRIPLE_QUOTE = re.compile("(\"\"\"|''')")
"""A regular expression matching the triple quotes used to delimit python docstrings."""

# fmt: off
# Order matters!
DEFAULT_MODULES = [
    "esbonio.lsp.directives",      # Generic directive support
    "esbonio.lsp.roles",           # Generic roles support
    "esbonio.lsp.rst.directives",  # Specialised support for docutils directives
    "esbonio.lsp.rst.roles",       # Specialised support for docutils roles
]
"""The modules to load in the default configuration of the server."""
# fmt: on


class CompletionContext:
    """Captures the context within which a completion request has been made."""

    def __init__(
        self,
        *,
        doc: Document,
        location: str,
        match: "re.Match",
        position: Position,
        config: ServerCompletionConfig,
        capabilities: ClientCapabilities,
    ):
        self.doc: Document = doc
        """The document within which the completion request was made."""

        self.location: str = location
        """The location type where the request was made.
        See :meth:`~esbonio.lsp.rst.RstLanguageServer.get_location_type` for details."""

        self.match: "re.Match" = match
        """The match object describing the site of the completion request."""

        self.position: Position = position
        """The position at which the completion request was made."""

        self.config: ServerCompletionConfig = config
        """User supplied configuration options."""

        self._client_capabilities: ClientCapabilities = capabilities

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return (
            f"CompletionContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"
        )

    @property
    def commit_characters_support(self) -> bool:
        """Indicates if the client supports commit characters."""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.commit_characters_support",
            False,
        )

    @property
    def deprecated_support(self) -> bool:
        """Indicates if the client supports the deprecated field on a
        ``CompletionItem``."""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.deprecated_support",
            False,
        )

    @property
    def documentation_formats(self) -> List[MarkupKind]:
        """The list of documentation formats supported by the client."""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.documentation_format",
            [],
        )

    @property
    def insert_replace_support(self) -> bool:
        """Indicates if the client supports ``InsertReplaceEdit``."""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.insert_replace_support",
            False,
        )

    @property
    def preselect_support(self) -> bool:
        """Indicates if the client supports the preselect field on a
        ``CompletionItem``."""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.preselect_support",
            False,
        )

    @property
    def snippet_support(self) -> bool:
        """Indicates if the client supports snippets"""
        return get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.snippet_support",
            False,
        )

    @property
    def supported_tags(self) -> List[CompletionItemTag]:
        """The list of ``CompletionItemTags`` supported by the client."""
        capabilities = get_capability(
            self._client_capabilities,
            "text_document.completion.completion_item.tag_support",
            None,
        )

        if not capabilities:
            return []

        return capabilities.value_set


class DocumentLinkContext:
    """Captures the context within which a document link request has been made."""

    def __init__(self, *, doc: Document, capabilities: ClientCapabilities):
        self.doc = doc
        """The document within which the document link request was made."""

        self._client_capabilities = capabilities

    @property
    def tooltip_support(self) -> bool:
        """Indicates if the client supports tooltips."""
        return get_capability(
            self._client_capabilities,
            "text_document.document_link.tooltip_support",
            False,
        )


class DefinitionContext:
    """A class that captures the context within which a definition request has been
    made."""

    def __init__(
        self, *, doc: Document, location: str, match: "re.Match", position: Position
    ):
        self.doc = doc
        """The document within which the definition request was made."""

        self.location = location
        """The location type where the request was made.
        See :meth:`~esbonio.lsp.rst.RstLanguageServer.get_location_type` for details."""

        self.match = match
        """The match object describing the site of the definition request."""

        self.position = position
        """The position at which the definition request was made."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return (
            f"DefinitionContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"
        )


class ImplementationContext:
    """A class that captures the context within which an implementation request has been
    made."""

    def __init__(
        self, *, doc: Document, location: str, match: "re.Match", position: Position
    ):
        self.doc = doc
        """The document within which the implementation request was made."""

        self.location = location
        """The location type where the request was made.
        See :meth:`~esbonio.lsp.rst.RstLanguageServer.get_location_type` for details."""

        self.match = match
        """The match object describing the site of the implementation request."""

        self.position = position
        """The position at which the implementation request was made."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return f"ImplementationContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"


class HoverContext:
    """A class that captures the context within a hover request has been made."""

    def __init__(
        self,
        *,
        doc: Document,
        location: str,
        match: "re.Match",
        position: Position,
        capabilities: ClientCapabilities,
    ):
        self.doc = doc
        self.location = location
        self.match = match
        self.position = position
        self._client_capabilities = capabilities

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return f"HoverContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"

    @property
    def content_formats(self) -> List[MarkupKind]:
        """The list of content formats supported by the client."""
        return get_capability(
            self._client_capabilities, "text_document.hover.content_format", []
        )


class LanguageFeature:
    """Base class for language features."""

    def __init__(self, rst: "RstLanguageServer"):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def initialize(self, options: InitializeParams) -> None:
        """Called once when the server is first initialized."""

    def initialized(self, params: InitializedParams) -> None:
        """Called once upon receipt of the `initialized` notification from the client."""

    def on_shutdown(self, *args):
        """Called as the server is shutting down."""

    def save(self, params: DidSaveTextDocumentParams) -> None:
        """Called each time a document is saved."""

    def delete_files(self, params: DeleteFilesParams) -> None:
        """Called each time files are deleted."""

    def code_action(self, params: CodeActionParams) -> List[CodeAction]:
        """Called when code actions should be computed."""
        return []

    hover_triggers: List["re.Pattern"] = []

    def hover(self, context: HoverContext) -> str:
        """Called when request textDocument/hover is sent.

        This method shall return the contents value of textDocument/hover response.

        """
        return ""

    completion_triggers: List["re.Pattern"] = []
    """A list of regular expressions used to determine if the
    :meth`~esbonio.lsp.rst.LanguageFeature.complete` method should be called on the
    current line."""

    def complete(self, context: CompletionContext) -> List[CompletionItem]:
        """Called if any of the given ``completion_triggers`` match the current line.

        This method should return a list of ``CompletionItem`` objects.

        Parameters
        ----------
        context:
           The context of the completion request
        """
        return []

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:
        """Called with a completion item the user has selected in the UI,
        allowing for additional details to be provided e.g. documentation.

        Parameters
        ----------
        item:
            The completion item to provide additional details for.
        """
        return item

    definition_triggers: List["re.Pattern"] = []
    """A list of regular expressions used to determine if the
    :meth:`~esbonio.lsp.rst.LanguageFeature.definition` method should be called."""

    def definition(self, context: DefinitionContext) -> List[Location]:
        """Called if any of the given ``definition_triggers`` match the current line.

        This method should return a list of ``Location`` objects.

        Parameters
        ----------
        context:
           The context of the definition request.
        """
        return []

    implementation_triggers: List["re.Pattern"] = []
    """A list of regular expressions used to determine if the
    :meth:`~esbonio.lsp.rst.LanguageFeature.implementation` method should be called."""

    def implementation(self, context: ImplementationContext) -> List[Location]:
        """Called if any of the given ``implementation_triggers`` match the current line.

        This method should return a list of ``Location`` objects.

        Parameters
        ----------
        context:
           The context of the implementation request.
        """
        return []

    def document_link(self, context: DocumentLinkContext) -> List[DocumentLink]:
        """Called whenever a ``textDocument/documentLink`` request is received."""
        return []


class DiagnosticList(collections.UserList):
    """A list type dedicated to holding diagnostics.

    This is mainly to ensure that only one instance of a diagnostic ever gets
    reported.
    """

    def append(self, item: Diagnostic):
        if not isinstance(item, Diagnostic):
            raise TypeError("Expected Diagnostic")

        for existing in self.data:
            fields = [
                existing.range == item.range,
                existing.message == item.message,
                existing.severity == item.severity,
                existing.code == item.code,
                existing.source == item.source,
            ]

            if all(fields):
                # Item already added, nothing to do.
                return

        self.data.append(item)


class RstLanguageServer(LanguageServer):
    """A generic reStructuredText language server."""

    def __init__(self, logger: Optional[logging.Logger] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = logger or logging.getLogger(__name__)
        """The base logger that should be used for all language sever log entries."""

        self.converter = self.lsp._converter
        """The cattrs converter instance we should use."""

        self.user_config: InitializationOptions = InitializationOptions()
        """The user's configuration."""

        self._diagnostics: Dict[Tuple[str, str], List[Diagnostic]] = {}
        """Where we store and manage diagnostics."""

        self._loaded_extensions: Dict[str, Any] = {}
        """Record of modules that have been loaded."""

        self._features: Dict[str, LanguageFeature] = {}
        """The collection of language features registered with the server."""

    @property
    def configuration(self) -> Dict[str, Any]:
        """Return the server's actual configuration."""
        return converter.unstructure(self.user_config)

    def initialize(self, params: InitializeParams):
        self.user_config = converter.structure(
            params.initialization_options or {}, InitializationOptions
        )
        setup_logging(self, self.user_config.server)

    def initialized(self, params: InitializedParams):
        pass

    def on_shutdown(self, *args):
        pass

    def save(self, params: DidSaveTextDocumentParams):
        pass

    def delete_files(self, params: DeleteFilesParams):
        pass

    def build(self, force_all: bool = False, filenames: Optional[List[str]] = None):
        pass

    def load_extension(self, name: str, setup: Callable):
        """Load the given setup function as an extension.

        If an extension with the given ``name`` already exists, the given setup function
        will be ignored.

        The ``setup`` function can declare dependencies in the form of type
        annotations.

        .. code-block:: python

           from esbonio.lsp.roles import Roles
           from esbonio.lsp.sphinx import SphinxLanguageServer

           def esbonio_setup(rst: SphinxLanguageServer, roles: Roles):
               ...

        In this example the setup function is requesting instances of the
        :class:`~esbonio.lsp.sphinx.SphinxLanguageServer` and the
        :class:`~esbonio.lsp.roles.Roles` language feature.

        Parameters
        ----------
        name
           The name to give the extension

        setup
           The setup function to call
        """

        if name in self._loaded_extensions:
            self.logger.debug("Skipping extension '%s', already loaded", name)
            return

        arguments = _get_setup_arguments(self, setup, name)
        if not arguments:
            return

        try:
            setup(**arguments)

            self.logger.debug("Loaded extension '%s'", name)
            self._loaded_extensions[name] = setup
        except Exception:
            self.logger.error(
                "Unable to load extension '%s'\n%s", name, traceback.format_exc()
            )

    def add_feature(self, feature: "LanguageFeature"):
        """Register a language feature with the server.

        Parameters
        ----------
        feature
           The language feature
        """

        key = f"{feature.__module__}.{feature.__class__.__name__}"
        self._features[key] = feature

    @typing.overload
    def get_feature(self, key: str) -> "Optional[LanguageFeature]":
        ...

    @typing.overload
    def get_feature(self, key: Type[LF]) -> Optional[LF]:
        ...

    def get_feature(self, key):
        """Returns the requested language feature if it exists, otherwise it returns
        ``None``.

        Parameters
        ----------
        key: str | Type[LanguageFeature]
           A feature can be referenced either by its class definition (preferred) or by
           a string representing the language feature's dotted name e.g.
           ``a.b.c.ClassName``.

           .. deprecated:: 0.14.0

              Passing a string ``key`` to this method is deprecated and will become an
              error in ``v1.0``.
        """

        if isinstance(key, str):
            warnings.warn(
                "Language features should be referenced by their class definition, "
                "this will become an error in v1.0.",
                DeprecationWarning,
                stacklevel=2,
            )

        elif issubclass(key, LanguageFeature):
            key = f"{key.__module__}.{key.__name__}"

        else:
            raise TypeError("Expected language feature definition")

        return self._features.get(key, None)

    def get_doctree(
        self, *, docname: Optional[str] = None, uri: Optional[str] = None
    ) -> Optional[Any]:
        # Not currently implemented for vanilla docutils projects.
        return None

    def get_initial_doctree(self, uri: str) -> Optional[Any]:
        """Return the initial doctree corresponding to the specified document.

        An "initial" doctree can be thought of as the abstract syntax tree of a
        reStructuredText document. This method disables all role and directives
        from being executed, instead they are replaced with nodes that simply
        represent that they exist.

        Parameters
        ----------
        uri
           Returns the doctree that corresponds with the given uri.
        """

        doc = self.workspace.get_document(uri)
        loctype = self.get_location_type(doc, Position(line=1, character=0))

        if loctype != "rst":
            return None

        self.logger.debug("Getting initial doctree for: '%s'", Uri.to_fs_path(uri))

        try:
            return read_initial_doctree(doc, self.logger)
        except Exception:
            self.logger.error(traceback.format_exc())
            return None

    def get_directives(self) -> Dict[str, Directive]:
        """Return a dictionary of the known directives

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``.
           Use the :meth:`~esbonio.lsp.directives.Directives.get_directives` method
           instead.
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_directives() is deprecated and will be removed in v1.0."
            " Instead call the get_directives() method on the Directives language "
            "feature.",
            DeprecationWarning,
            stacklevel=2,
        )

        feature = self.get_feature("esbonio.lsp.directives.Directives")
        if feature is None:
            return {}

        return feature.get_directives()  # type: ignore

    def get_directive_options(self, name: str) -> Dict[str, Any]:
        """Return the options specification for the given directive.

        .. deprecated:: 0.14.2

           This will be removed in ``v1.0``
        """

        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_directive_options() is deprecated and will be removed in "
            "v1.0.",
            DeprecationWarning,
            stacklevel=2,
        )

        directive = self.get_directives().get(name, None)
        if directive is None:
            return {}

        return directive.option_spec or {}

    def get_roles(self) -> Dict[str, Any]:
        """Return a dictionary of known roles.

        .. deprecated:: 0.15.0

           This will be removed in ``v1.0``.
           Use the :meth:`~esbonio.lsp.roles.Roles.get_roles` method instead.
        """
        clsname = self.__class__.__name__
        warnings.warn(
            f"{clsname}.get_roles() is deprecated and will be removed in v1.0. "
            "Instead call the get_roles() method on the Roles language "
            "feature.",
            DeprecationWarning,
            stacklevel=2,
        )

        feature = self.get_feature("esbonio.lsp.roles.Roles")
        if feature is None:
            return {}

        return feature.get_roles()  # type: ignore

    def get_default_role(self) -> Tuple[Optional[str], Optional[str]]:
        """Return the default role for the project."""
        return None, None

    def clear_diagnostics(self, source: str, uri: Optional[str] = None) -> None:
        """Clear diagnostics from the given source.

        Parameters
        ----------
        source:
           The source from which to clear diagnostics.
        uri:
           If given, clear diagnostics from within just this uri. Otherwise, all
           diagnostics from the given source are cleared.
        """

        if uri:
            uri = normalise_uri(uri)

        for key in self._diagnostics.keys():
            clear_source = source == key[0]
            clear_uri = uri == key[1] or uri is None

            if clear_source and clear_uri:
                self._diagnostics[key] = []

    def add_diagnostics(self, source: str, uri, diagnostic: Diagnostic):
        """Add a diagnostic to the given source and uri.

        Parameters
        ----------
        source
           The source the diagnostics are from
        uri
           The uri the diagnostics are associated with
        diagnostic
           The diagnostic to add
        """
        key = (source, normalise_uri(uri))
        self._diagnostics.setdefault(key, []).append(diagnostic)

    def set_diagnostics(
        self, source: str, uri: str, diagnostics: List[Diagnostic]
    ) -> None:
        """Set the diagnostics for the given source and uri.

        Parameters
        ----------
        source:
           The source the diagnostics are from
        uri:
           The uri the diagnostics are associated with
        diagnostics:
           The diagnostics themselves
        """
        uri = normalise_uri(uri)
        self._diagnostics[(source, uri)] = diagnostics

    def sync_diagnostics(self) -> None:
        """Update the client with the currently stored diagnostics."""

        uris = {uri for _, uri in self._diagnostics.keys()}
        diagnostics = {uri: DiagnosticList() for uri in uris}

        for (source, uri), diags in self._diagnostics.items():
            for diag in diags:
                diag.source = source
                diagnostics[uri].append(diag)

        for uri, diag_list in diagnostics.items():
            self.logger.debug("Publishing %d diagnostics for: %s", len(diag_list), uri)
            self.publish_diagnostics(uri, diag_list.data)

    def get_location_type(self, document: Document, position: Position) -> str:
        """Given a document and a position, return the type of location.

        Returns one of the following values:

        - ``rst``: Indicates that the position is within an reStructuredText document
        - ``py``: Indicates that the position is within code in a Python file
        - ``docstring``: Indicates that the position is within a docstring in a
          Python file.

        If the location type cannot be determined, this function will fall back to
        ``rst``.

        Parameters
        ----------
        doc
           The document associated with the given position

        position
           The position to determine the type of
        """
        doc = self.workspace.get_document(document.uri)
        fpath = pathlib.Path(Uri.to_fs_path(doc.uri))

        # Prefer the document's language_id, but fallback to file extensions
        loctype = getattr(doc, "language_id", None) or fpath.suffix

        if loctype in {".rst", "rst", "restructuredtext"}:
            return "rst"

        if loctype in {".py", "py", "python"}:
            # Let's count how many pairs of triple quotes are above us in the file
            # even => we're outside a docstring
            # odd  => we're within a docstring
            source = self.text_to_position(doc, position)
            count = len(TRIPLE_QUOTE.findall(source))
            return "py" if count % 2 == 0 else "docstring"

        # Fallback to rst
        self.logger.debug("Unable to determine location type for uri: %s", doc.uri)
        return "rst"

    def line_at_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the line corresponding to the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position
        position:
           The position representing the line to retrieve
        """

        try:
            return doc.lines[position.line]
        except IndexError:
            return ""

    def line_to_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the line up until the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position.
        position:
           The position representing the line to retrieve.
        """

        line = self.line_at_position(doc, position)
        return line[: position.character]

    def preview(self, options: Dict[str, Any]) -> Dict[str, Any]:
        """Generate a preview of the documentation."""
        name = self.__class__.__name__
        self.show_message(
            f"Previews are not currently supported by {name} based servers"
        )

        return {}

    def text_to_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the document up until the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position
        position:
           The position representing the point at which to stop gathering text.
        """
        idx = doc.offset_at_position(position)
        return doc.source[:idx]


def normalise_uri(uri: str) -> str:
    uri = Uri.from_fs_path(Uri.to_fs_path(uri))

    # Paths on windows are case insensitive.
    if IS_WIN:
        uri = uri.lower()

    return uri


def _get_setup_arguments(
    server: RstLanguageServer, setup: Callable, modname: str
) -> Optional[Dict[str, Any]]:
    """Given a setup function, try to construct the collection of arguments to pass to
    it.
    """
    annotations = typing.get_type_hints(setup)
    parameters = {
        p.name: annotations[p.name]
        for p in inspect.signature(setup).parameters.values()
    }

    args = {}
    for name, type_ in parameters.items():
        if issubclass(server.__class__, type_):
            args[name] = server
            continue

        if issubclass(type_, LanguageFeature):
            # Try and obtain an instance of the requested language feature.
            feature = server.get_feature(type_)
            if feature is not None:
                args[name] = feature
                continue

            server.logger.debug(
                "Skipping extension '%s', server missing requested feature: '%s'",
                modname,
                type_,
            )
            return None

        server.logger.error(
            "Skipping extension '%s', parameter '%s' has unsupported type: '%s'",
            modname,
            name,
            type_,
        )
        return None

    return args


cli = setup_cli("esbonio.lsp.rst", "Esbonio's reStructuredText language server.")
cli.set_defaults(modules=DEFAULT_MODULES)
cli.set_defaults(server_cls=RstLanguageServer)



"""Support for code-blocks and related directives."""
import textwrap
from typing import List

from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionItemKind
from lsprotocol.types import MarkupContent
from lsprotocol.types import MarkupKind
from pygments.lexers import get_all_lexers

from esbonio.lsp.directives import Directives
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.sphinx import SphinxLanguageServer


class CodeBlocks:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)
        self._lexers: List[CompletionItem] = []

    @property
    def lexers(self) -> List[CompletionItem]:
        if len(self._lexers) == 0:
            self._index_lexers()

        return self._lexers

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:
        if name in {"code-block", "highlight"}:
            return self.lexers

        return []

    def _index_lexers(self):
        self._lexers = []

        for name, labels, files, mimes in get_all_lexers():
            for label in labels:
                documentation = f"""\
                ### {name}

                Filenames:  {', '.join(files)}

                MIME Types: {', '.join(mimes)}
                """

                item = CompletionItem(
                    label=label,
                    kind=CompletionItemKind.Constant,
                    documentation=MarkupContent(
                        kind=MarkupKind.Markdown, value=textwrap.dedent(documentation)
                    ),
                )

                self._lexers.append(item)


def esbonio_setup(rst: SphinxLanguageServer, directives: Directives):
    directives.add_argument_completion_provider(CodeBlocks(rst))



from typing import Iterable
from typing import Optional

from esbonio.lsp import CompletionContext
from esbonio.lsp.directives import DirectiveLanguageFeature
from esbonio.lsp.directives import Directives
from esbonio.lsp.sphinx import SphinxLanguageServer


class AutoDoc(DirectiveLanguageFeature):
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst

    def suggest_options(
        self, context: CompletionContext, directive: str, domain: Optional[str]
    ) -> Iterable[str]:
        if self.rst.app is None or not directive.startswith("auto"):
            return []

        # The autoxxxx set of directives need special support as their options are
        # stored on "documenters" instead of the directive implementation itself.
        name = directive.replace("auto", "")
        documenter = self.rst.app.registry.documenters.get(name, None)

        if documenter is None:
            self.rst.logger.debug(
                "Unable to find documenter for directive: '%s'", directive
            )
            return []

        return documenter.option_spec.keys()


def esbonio_setup(rst: SphinxLanguageServer, directives: Directives):
    directives.add_feature(AutoDoc(rst))



from __future__ import annotations

import asyncio
import inspect
import typing
from typing import Callable
from typing import Dict
from typing import Optional

import lsprotocol.types as lsp

from esbonio.server import LanguageFeature
from esbonio.server import Uri

from .config import SphinxConfig

if typing.TYPE_CHECKING:
    from .client import SphinxClient


SphinxClientFactory = Callable[["SphinxManager"], "SphinxClient"]


class SphinxManager(LanguageFeature):
    """Responsible for managing Sphinx application instances."""

    def __init__(self, client_factory: SphinxClientFactory, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.client_factory = client_factory
        """Used to create new Sphinx client instances."""

        self.clients: Dict[Uri, SphinxClient] = {}
        """Holds currently active Sphinx clients."""

        self.handlers: Dict[str, set] = {}
        """Collection of handlers for various events."""

        self._pending_builds: Dict[str, asyncio.Task] = {}
        """Holds tasks that will trigger a build after a given delay if not cancelled."""

        self._client_creating: Optional[asyncio.Task] = None
        """If set, indicates we're in the process of setting up a new client."""

    def add_listener(self, event: str, handler):
        self.handlers.setdefault(event, set()).add(handler)

    async def document_change(self, params: lsp.DidChangeTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None or client.id is None:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        self._pending_builds[client.id] = asyncio.create_task(
            self.trigger_build_after(uri, client.id, delay=2)
        )

    async def document_open(self, params: lsp.DidOpenTextDocumentParams):
        # Ensure that a Sphinx app instance is created the first time a document in a
        # given project is opened.
        if (uri := Uri.parse(params.text_document.uri)) is not None:
            await self.get_client(uri)

    async def document_save(self, params: lsp.DidSaveTextDocumentParams):
        if (uri := Uri.parse(params.text_document.uri)) is None:
            return

        client = await self.get_client(uri)
        if client is None or client.id is None:
            return

        # Cancel any existing pending builds
        if (task := self._pending_builds.pop(client.id, None)) is not None:
            task.cancel()

        await self.trigger_build(uri)

    async def trigger_build_after(self, uri: Uri, app_id: str, delay: float):
        """Trigger a build for the given uri after the given delay."""
        await asyncio.sleep(delay)

        self._pending_builds.pop(app_id)
        self.logger.debug("Triggering build")
        await self.trigger_build(uri)

    async def trigger_build(self, uri: Uri):
        """Trigger a build for the relevant Sphinx application for the given uri."""
        client = await self.get_client(uri)
        if client is None or client.building:
            return

        # Pass through any unsaved content to the Sphinx agent.
        content_overrides: Dict[str, str] = {}
        for src_uri in client.build_file_map.keys():
            doc = self.server.workspace.get_document(str(src_uri))
            doc_version = doc.version or 0
            saved_version = getattr(doc, "saved_version", 0)

            if saved_version < doc_version and (fs_path := src_uri.fs_path) is not None:
                content_overrides[fs_path] = doc.source

        result = await client.build(content_overrides=content_overrides)

        # Update diagnostics
        source = f"sphinx[{client.id}]"
        self.server.clear_diagnostics(source)
        for uri, items in client.diagnostics.items():
            diagnostics = [
                lsp.Diagnostic(
                    range=d.range,  # type: ignore[arg-type]
                    message=d.message,
                    source=source,
                    severity=d.severity,  # type: ignore[arg-type]
                )
                for d in items
            ]
            self.server.set_diagnostics(f"sphinx[{client.id}]", uri, diagnostics)

        self.server.sync_diagnostics()

        # Notify listeners.
        for listener in self.handlers.get("build", set()):
            try:
                # TODO: Concurrent awaiting?
                res = listener(client, result)
                if inspect.isawaitable(res):
                    await res
            except Exception:
                name = f"{listener}"
                self.logger.error("Error in build handler '%s'", name, exc_info=True)

    async def get_client(self, uri: Uri) -> Optional[SphinxClient]:
        """Given a uri, return the relevant sphinx client instance for it."""

        # Wait until the new client is created - it might be the one we're looking for!
        if self._client_creating:
            await self._client_creating

        # Always check the fully resolved uri.
        resolved_uri = uri.resolve()

        for src_uri, client in self.clients.items():
            if resolved_uri in client.build_file_map:
                return client

            # For now assume a single client instance per srcdir.
            # This *should* prevent us from spwaning endless client instances
            # when given a file located near a valid Sphinx project - but not actually
            # part of it.
            in_src_dir = str(resolved_uri).startswith(str(src_uri))
            if in_src_dir:
                # Of course, we can only tell if a uri truly is not in a project
                # when the build file map is populated!
                if len(client.build_file_map) == 0:
                    return client

                return None

        # Create a new client instance.
        self._client_creating = asyncio.create_task(self._create_client(uri))
        return await self._client_creating

    async def _create_client(self, uri: Uri) -> Optional[SphinxClient]:
        """Create a new sphinx client instance."""
        config = await self.server.get_user_config(
            "esbonio.sphinx", SphinxConfig, scope=uri
        )
        if config is None:
            return None

        resolved = config.resolve(uri, self.server.workspace, self.logger)
        if resolved is None:
            return None

        client = self.client_factory(self)
        await client.start(resolved)

        sphinx_info = await client.create_application(resolved)

        if client.src_uri is None:
            self.logger.error("No src uri!")
            await client.stop()
            return None

        self.server.lsp.notify("sphinx/appCreated", sphinx_info)
        self.clients[client.src_uri] = client
        return client



import importlib.resources


def read_string(package: str, filename: str) -> str:
    """Light wrapper around ``importlib.resources`` that should work across Python
    versions.

    Parameters
    ----------
    package
       The module/package to read from

    filename
       The file within ``package`` to read

    Returns
    -------
    str
       The contents of the specified text file.
    """

    # `files` only available in Python 3.9+
    if hasattr(importlib.resources, "files"):
        with importlib.resources.files(package).joinpath(filename).open("r") as f:
            return f.read()

    # `open_text` deprecated in Python 3.11, so let's only rely on it when we
    # have to.
    with importlib.resources.open_text(package, filename) as f:
        return f.read()



from __future__ import annotations

import enum
import json
import logging
import pathlib
import re
import textwrap
import traceback
import typing
from typing import List
from typing import Tuple

from lsprotocol.types import Diagnostic
from lsprotocol.types import DiagnosticSeverity
from lsprotocol.types import DiagnosticTag
from lsprotocol.types import Position
from lsprotocol.types import Range

from ._uri import Uri

if typing.TYPE_CHECKING:
    from .server import EsbonioLanguageServer
    from .server import ServerConfig


LOG_NAMESPACE = "esbonio"
LOG_LEVELS = {
    "debug": logging.DEBUG,
    "error": logging.ERROR,
    "info": logging.INFO,
}


# e.g. /.../filename.rst:54: (ERROR/3) Unexpected indentation.
#      c:\...\filename.rst:54: (ERROR/3) Unexpected indentation.
DOCUTILS_ERROR = re.compile(
    r"""
    ^\s*(?P<filepath>.+?):
    (?P<linum>\d+):
    \s*\((?P<levelname>\w+)/(?P<levelnum>\d)\)
    (?P<message>.*)$
    """,
    re.VERBOSE,
)

DOCUTILS_SEVERITY = {
    0: DiagnosticSeverity.Hint,
    1: DiagnosticSeverity.Information,
    2: DiagnosticSeverity.Warning,
    3: DiagnosticSeverity.Error,
    4: DiagnosticSeverity.Error,
}


class LogFilter(logging.Filter):
    """A log filter that accepts message from any of the listed logger names."""

    def __init__(self, names):
        self.names = names

    def filter(self, record):
        return any(record.name == name for name in self.names)


class MemoryHandler(logging.Handler):
    """A logging handler that caches messages in memory."""

    def __init__(self):
        super().__init__()
        self.records: List[logging.LogRecord] = []

    def emit(self, record: logging.LogRecord) -> None:
        self.records.append(record)


class LspHandler(logging.Handler):
    """A logging handler that will send log records to an LSP client."""

    def __init__(
        self, server: EsbonioLanguageServer, show_deprecation_warnings: bool = False
    ):
        super().__init__()
        self.server = server
        self.show_deprecation_warnings = show_deprecation_warnings

    def get_warning_path(self, warning: str) -> Tuple[str, List[str]]:
        """Determine the filepath that the warning was emitted from."""

        path, *parts = warning.split(":")

        # On windows the rest of the path will be in the first element of parts.
        if pathlib.Path(warning).drive:
            path += f":{parts.pop(0)}"

        return path, parts

    def handle_warning(self, record: logging.LogRecord):
        """Publish warnings to the client as diagnostics."""

        if not isinstance(record.args, tuple):
            self.server.logger.debug(
                "Unable to handle warning, expected tuple got: %s", record.args
            )
            return

        # The way warnings are logged is different in Python 3.11+
        if len(record.args) == 0:
            argument = record.msg
        else:
            argument = record.args[0]  # type: ignore

        if not isinstance(argument, str):
            self.server.logger.debug(
                "Unable to handle warning, expected string got: %s", argument
            )
            return

        warning, *_ = argument.split("\n")
        path, (linenum, category, *msg) = self.get_warning_path(warning)

        category = category.strip()
        message = ":".join(msg).strip()

        try:
            line = int(linenum)
        except ValueError:
            line = 1
            self.server.logger.debug(
                "Unable to parse line number: '%s'\n%s", linenum, traceback.format_exc()
            )

        tags = []
        if category == "DeprecationWarning":
            tags.append(DiagnosticTag.Deprecated)

        diagnostic = Diagnostic(
            range=Range(
                start=Position(line=line - 1, character=0),
                end=Position(line=line, character=0),
            ),
            message=message,
            severity=DiagnosticSeverity.Warning,
            tags=tags,
        )

        self.server.add_diagnostics("esbonio", Uri.for_file(path), diagnostic)
        self.server.sync_diagnostics()

    def handle_diagnostic(self, record: logging.LogRecord):
        """Look for any diagnostics to report in the log message."""

        if (match := DOCUTILS_ERROR.match(record.msg)) is not None:
            uri = Uri.for_file(match.group("filepath"))
            line = int(match.group("linum"))
            severity = int(match.group("levelnum"))

            diagnostic = Diagnostic(
                message=match.group("message").strip(),
                severity=DOCUTILS_SEVERITY.get(severity),
                range=Range(
                    start=Position(line=line - 1, character=0),
                    end=Position(line=line, character=0),
                ),
            )
            self.server.add_diagnostics("docutils", uri, diagnostic)

    def emit(self, record: logging.LogRecord) -> None:
        """Sends the record to the client."""

        # To avoid infinite recursions, it's simpler to just ignore all log records
        # coming from pygls...
        if "pygls" in record.name:
            return

        if record.name == "py.warnings":
            if not self.show_deprecation_warnings:
                return

            self.handle_warning(record)
        else:
            self.handle_diagnostic(record)

        log = self.format(record).strip()
        self.server.show_message_log(log)


def setup_logging(server: EsbonioLanguageServer, config: ServerConfig):
    """Setup logging to route log messages to the language client as
    ``window/logMessage`` messages.

    Parameters
    ----------
    server
       The server to use to send messages

    config
       The configuration to use
    """

    level = LOG_LEVELS[config.log_level]

    warnlog = logging.getLogger("py.warnings")
    logger = logging.getLogger(LOG_NAMESPACE)
    logger.setLevel(level)

    lsp_handler = LspHandler(server, config.show_deprecation_warnings)
    lsp_handler.setLevel(level)

    if len(config.log_filter) > 0:
        lsp_handler.addFilter(LogFilter(config.log_filter))

    formatter = logging.Formatter("[%(name)s] %(message)s")
    lsp_handler.setFormatter(formatter)

    # Look to see if there are any cached messages we should forward to the client.
    for handler in logger.handlers:
        if not isinstance(handler, MemoryHandler):
            continue

        for record in handler.records:
            if logger.isEnabledFor(record.levelno):
                lsp_handler.emit(record)

        logger.removeHandler(handler)

    logger.addHandler(lsp_handler)
    warnlog.addHandler(lsp_handler)


def dump(obj) -> str:
    """Debug helper function that converts an object to JSON."""

    def default(o):
        if isinstance(o, enum.Enum):
            return o.value

        fields = {}
        for k, v in o.__dict__.items():
            if v is None:
                continue

            # Truncate long strings - but not uris!
            if isinstance(v, str) and not k.lower().endswith("uri"):
                v = textwrap.shorten(v, width=25)

            fields[k] = v

        return fields

    return json.dumps(obj, default=default, indent=2)



import logging
from functools import partial
from http.server import HTTPServer
from http.server import SimpleHTTPRequestHandler
from multiprocessing import Queue
from typing import Any
from typing import Type

try:
    from http.server import ThreadingHTTPServer

    ServerClass: Type[HTTPServer] = ThreadingHTTPServer
except ImportError:
    # ThreadingHTTPServer is only availble in Python 3.7+
    ServerClass = HTTPServer

logger = logging.getLogger(__name__)


class RequestHandler(SimpleHTTPRequestHandler):
    def log_message(self, format: str, *args: Any) -> None:
        return logger.debug(format, *args)


def make_preview_server(directory: str) -> HTTPServer:
    """Construst a http server that can be used to preview the docs."""
    handler_class = partial(RequestHandler, directory=directory)
    return ServerClass(("localhost", 0), handler_class)


def start_preview_server(q: Queue, directory: str):
    """Start a preview server in the given directory.

    The server's port number will be sent back via the given ``q`` object.
    """

    handler_class = partial(RequestHandler, directory=directory)
    server = ServerClass(("localhost", 0), handler_class)
    q.put(server.server_port)

    server.serve_forever()



from typing import Dict
from typing import List
from typing import Optional
from typing import Protocol

from esbonio.server import Uri
from esbonio.sphinx_agent import types

from .config import SphinxConfig


class SphinxClient(Protocol):
    """Describes the API language features can use to inspect/manipulate a Sphinx
    application instance."""

    @property
    def id(self) -> Optional[str]:
        """The id of the Sphinx instance."""

    @property
    def builder(self) -> Optional[str]:
        """The name of the Sphinx builder."""

    @property
    def building(self) -> bool:
        """Indicates if the Sphinx instance is building."""

    @property
    def build_uri(self) -> Optional[Uri]:
        """The URI to the Sphinx application's build dir."""

    @property
    def build_file_map(self) -> Dict[Uri, str]:
        """A mapping of source file uris to the corresponding build path that contains
        their content.

        Example
        -------
        >>> client.build_file_map
        {
           Uri(scheme='file', path='/path/to/index.rst'): "index.html",
        }
        """
        ...

    @property
    def diagnostics(self) -> Dict[Uri, List[types.Diagnostic]]:
        """A mapping of source file uris to any diagnostic items."""
        ...

    @property
    def conf_uri(self) -> Optional[Uri]:
        """The URI to the Sphinx application's conf dir."""

    @property
    def src_uri(self) -> Optional[Uri]:
        """The URI to the Sphinx application's src dir."""

    async def start(self, config: SphinxConfig):
        """Start the client."""
        ...

    async def create_application(self, config: SphinxConfig) -> types.SphinxInfo:
        """Create a new Sphinx application instance."""
        ...

    async def build(
        self,
        *,
        filenames: Optional[List[str]] = None,
        force_all: bool = False,
        content_overrides: Optional[Dict[str, str]] = None,
    ) -> types.BuildResult:
        """Trigger a Sphinx build."""
        ...

    async def stop(self):
        """Stop the client."""



import os.path
import pathlib
from typing import List
from typing import Optional
from typing import Tuple

import pygls.uris as Uri
from lsprotocol.types import CompletionItem
from lsprotocol.types import Location
from lsprotocol.types import Position
from lsprotocol.types import Range
from pygls.workspace import Document

from esbonio.lsp.directives import Directives
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import DefinitionContext
from esbonio.lsp.rst import DocumentLinkContext
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.filepaths import complete_sphinx_filepaths
from esbonio.lsp.util.filepaths import path_to_completion_item


class Images:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:
        if domain or name not in {"figure", "image"}:
            return []

        if not self.rst.app:
            return []

        srcdir = self.rst.app.srcdir
        partial = context.match.group("argument")
        base = os.path.dirname(Uri.to_fs_path(context.doc.uri))
        items = complete_sphinx_filepaths(srcdir, base, partial)  # type: ignore[arg-type]

        return [path_to_completion_item(context, p) for p in items]

    def find_definitions(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        if domain or directive not in {"figure", "image"}:
            return []

        uri = self.resolve_path(context.doc, argument)
        if not uri:
            return []

        return [
            Location(
                uri=uri,
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=1, character=0),
                ),
            )
        ]

    def resolve_link(
        self,
        context: DocumentLinkContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> Tuple[Optional[str], Optional[str]]:
        if domain or directive not in {"figure", "image"}:
            return None, None

        if argument.startswith("https://") or argument.startswith("http://"):
            return argument, None

        return self.resolve_path(context.doc, argument), None

    def resolve_path(self, doc: Document, argument: str) -> Optional[str]:
        if argument.startswith("/"):
            if not self.rst.app:
                return None

            basedir = pathlib.Path(self.rst.app.srcdir)

            # Remove the leading '/' otherwise is will wipe out the basedir when
            # concatenated
            argument = argument[1:]

        else:
            basedir = pathlib.Path(Uri.to_fs_path(doc.uri)).parent

        try:
            fullpath = basedir / argument
            fpath = fullpath.resolve()

            if fpath.exists():
                return Uri.from_fs_path(str(fpath))
        except Exception:
            self.logger.debug("Unable to resolve filepath '%s'", fullpath)

        return None


def esbonio_setup(rst: SphinxLanguageServer, directives: Directives):
    images = Images(rst)

    directives.add_argument_definition_provider(images)
    directives.add_argument_completion_provider(images)
    directives.add_argument_link_provider(images)



import inspect
import logging
import traceback
from typing import Optional

import pygls.uris as Uri
from lsprotocol.types import Location
from lsprotocol.types import Position
from lsprotocol.types import Range


def get_object_location(obj: object, logger: logging.Logger) -> Optional[Location]:
    """Given an object, attempt to find the location of its implementation.

    Parameters
    ----------
    obj
       The object to find the implementation of

    logger
       A logger object
    """

    try:
        file = inspect.getsourcefile(obj)  # type: ignore
        if file is None:
            return None

        source, line = inspect.getsourcelines(obj)  # type: ignore
        return Location(
            uri=Uri.from_fs_path(file),
            range=Range(
                start=Position(line=line - 1, character=0),
                end=Position(line=line + len(source), character=0),
            ),
        )

    except Exception:
        logger.debug(
            "Unable to get implementation location\n%s", traceback.format_exc()
        )
        return None



"""This module implements the websocket server used to communicate with preivew
 windows."""
import asyncio
import json
import logging
import socket
from typing import Optional

from pygls.protocol import JsonRPCProtocol
from pygls.protocol import default_converter
from pygls.server import Server
from pygls.server import WebSocketTransportAdapter
from websockets.server import serve

from esbonio.server import EsbonioLanguageServer


class WebviewServer(Server):
    """The webview server controlls the webpage hosting the preview.

    Used to implement automatic reloads and features like sync scrolling.
    """

    lsp: JsonRPCProtocol

    def __init__(self, logger: logging.Logger, *args, **kwargs):
        super().__init__(JsonRPCProtocol, default_converter, *args, **kwargs)
        self.logger = logger
        self.lsp._send_only_body = True
        self.port = None

        self._connected = False
        self._editor_in_control: Optional[asyncio.Task] = None
        self._view_in_control: Optional[asyncio.Task] = None

    @property
    def connected(self) -> bool:
        """Indicates when we have an active connection to the client."""
        return self._connected

    def feature(self, feature_name: str, options=None):
        return self.lsp.fm.feature(feature_name, options)

    def reload(self):
        """Reload the current view."""
        if self.connected:
            self.lsp.notify("view/reload", {})

    def scroll(self, line: int):
        """Called by the editor to scroll the current webview."""
        if not self.connected or self._view_in_control:
            return

        # If the editor is already in control, reset the cooldown
        if self._editor_in_control:
            self._editor_in_control.cancel()

        self._editor_in_control = asyncio.create_task(self.cooldown("editor"))
        self.lsp.notify("view/scroll", {"line": line})

    async def cooldown(self, name: str):
        """Create a cooldown."""
        await asyncio.sleep(1)

        # Unset the cooldown
        self.logger.debug("%s cooldown ended", name)
        setattr(self, f"_{name}_in_control", None)

    async def start_ws(self, host: str, port: int) -> None:  # type: ignore[override]
        """Start the server."""

        async def connection(websocket):
            loop = asyncio.get_running_loop()
            transport = WebSocketTransportAdapter(websocket, loop)

            self.lsp.connection_made(transport)  # type: ignore[arg-type]
            self._connected = True
            self.logger.debug("Connected")

            async for message in websocket:
                self.lsp._procedure_handler(
                    json.loads(message, object_hook=self.lsp._deserialize_message)
                )

            self.logger.debug("Connection lost")
            self._connected = False

        async with serve(
            connection,
            host,
            port,
            # logger=self.logger.getChild("ws"),
            family=socket.AF_INET,  # Use IPv4 only.
        ) as ws_server:
            sock = list(ws_server.sockets)[0]
            self.port = sock.getsockname()[1]
            await asyncio.Future()  # run forever


def make_ws_server(
    esbonio: EsbonioLanguageServer, logger: logging.Logger
) -> WebviewServer:
    server = WebviewServer(logger)

    @server.feature("editor/scroll")
    def on_scroll(ls: WebviewServer, params):
        """Called by the webview to scroll the editor."""
        if not server.connected or server._editor_in_control:
            return

        # If the view is already in control, reset the cooldown.
        if server._view_in_control:
            server._view_in_control.cancel()

        server._view_in_control = asyncio.create_task(server.cooldown("view"))
        esbonio.lsp.notify("editor/scroll", dict(line=params.line))

    return server



"""Additional command line utilities for the SphinxLanguageServer."""
import argparse
import json
import sys
from typing import List

from esbonio.cli import esbonio_converter
from esbonio.lsp.sphinx import SphinxConfig


def config_cmd(args, extra):
    if args.to_cli:
        config_to_cli(args.to_cli)
        return 0

    return cli_to_config(extra)


def config_to_cli(config: str):
    converter = esbonio_converter()
    conf = converter.structure(json.loads(config), SphinxConfig)
    print(" ".join(conf.to_cli_args()))
    return 0


def cli_to_config(cli_args: List[str]):
    conf = SphinxConfig.from_arguments(cli_args=cli_args)
    if conf is None:
        return 1

    converter = esbonio_converter()
    print(json.dumps(converter.unstructure(conf), indent=2))
    return 0


cli = argparse.ArgumentParser(
    prog="esbonio-sphinx",
    description="Supporting commands and utilities for the SphinxLanguageServer.",
)
commands = cli.add_subparsers(title="commands")
config = commands.add_parser(
    "config",
    usage="%(prog)s [--from-cli] -- ARGS",
    description="configuration options helper.",
)
config.set_defaults(run=config_cmd)

mode = config.add_mutually_exclusive_group()
mode.add_argument(
    "--from-cli",
    action="store_true",
    default=True,
    help="convert sphinx-build cli options to esbonio's initialization options.",
)
mode.add_argument(
    "--to-cli",
    help="convert esbonio's initialization options to sphinx-build options",
)


def main():
    try:
        idx = sys.argv.index("--")
        args, extra = sys.argv[1:idx], sys.argv[idx + 1 :]
    except ValueError:
        args, extra = sys.argv[1:], None

    parsed_args = cli.parse_args(args)

    if hasattr(parsed_args, "run"):
        return parsed_args.run(parsed_args, extra)

    cli.print_help()


if __name__ == "__main__":
    sys.exit(main())



import logging
from typing import List
from typing import Optional

from docutils import nodes
from docutils.nodes import NodeVisitor
from lsprotocol.types import DocumentSymbol
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import SymbolKind


class SymbolVisitor(NodeVisitor):
    """A visitor used to build the hierarchy we return from a
    ``textDocument/documentSymbol`` request.

    """

    def __init__(self, logger: Optional[logging.Logger] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = logger or logging.getLogger(__name__)

        self.symbols: List[DocumentSymbol] = []
        """Holds the top-level symbols for the document"""

        self.symbol_stack: List[DocumentSymbol] = []
        """Used to maintain the current symbol hierarchy."""

    @property
    def current_symbol(self) -> Optional[DocumentSymbol]:
        if len(self.symbol_stack) == 0:
            return None

        return self.symbol_stack[-1]

    def push_symbol(self):
        symbol = DocumentSymbol(
            name="",
            kind=SymbolKind.String,
            range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            selection_range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            children=[],
        )
        current_symbol = self.current_symbol

        if not current_symbol:
            self.symbols.append(symbol)
        else:
            if current_symbol.children is None:
                current_symbol.children = [symbol]
            else:
                current_symbol.children.append(symbol)

        self.symbol_stack.append(symbol)
        return symbol

    def pop_symbol(self):
        self.symbol_stack.pop()

    def visit_section(self, node: nodes.Node) -> None:
        self.push_symbol()

    def depart_section(self, node: nodes.Node) -> None:
        self.pop_symbol()

    def visit_title(self, node: nodes.Node) -> None:
        symbol = self.current_symbol
        has_parent = True

        if symbol is None:
            has_parent = False
            symbol = self.push_symbol()

        name = node.astext()
        line = (node.line or 1) - 1

        symbol.name = name
        symbol.range.start.line = line
        symbol.range.end.line = line
        symbol.range.end.character = len(name) - 1
        symbol.selection_range.start.line = line
        symbol.selection_range.end.line = line
        symbol.selection_range.end.character = len(name) - 1

        if not has_parent:
            self.pop_symbol()

    def depart_title(self, node: nodes.Node) -> None:
        pass

    def visit_a_directive(self, node: nodes.Element):
        symbol = self.push_symbol()

        name = node["text"]  # type: ignore
        line = (node.line or 1) - 1

        symbol.name = name
        symbol.kind = SymbolKind.Class
        symbol.range.start.line = line
        symbol.range.end.line = line
        symbol.range.end.character = len(name) - 1
        symbol.selection_range.start.line = line
        symbol.selection_range.end.line = line
        symbol.selection_range.end.character = len(name) - 1

    def depart_a_directive(self, node: nodes.Node):
        self.pop_symbol()

    # TODO: Enable symbols for roles
    #       However the reported line numbers can be inaccurate...
    def visit_a_role(self, node: nodes.Node) -> None:
        ...

    def depart_a_role(self, node: nodes.Node) -> None:
        ...

    # TODO: Enable symbols for definition list items
    #       However the reported line numbers appear to be inaccurate...

    def visit_Text(self, node: nodes.Node) -> None:
        pass

    def depart_Text(self, node: nodes.Node) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass



from __future__ import annotations

import typing
from typing import List
from typing import Optional

from lsprotocol import types

if typing.TYPE_CHECKING:
    from .server import EsbonioLanguageServer


class LanguageFeature:
    """Base class for language features."""

    def __init__(self, server: EsbonioLanguageServer):
        self.server = server
        self.converter = server.converter
        self.logger = server.logger.getChild(self.__class__.__name__)

    def document_change(self, params: types.DidChangeTextDocumentParams):
        """Called when a text document is changed."""

    def document_close(self, params: types.DidCloseTextDocumentParams):
        """Called when a text document is closed."""

    def document_open(self, params: types.DidOpenTextDocumentParams):
        """Called when a text document is opened."""

    def document_save(self, params: types.DidSaveTextDocumentParams):
        """Called when a text document is saved."""

    def document_symbol(
        self, params: types.DocumentSymbolParams
    ) -> Optional[List[types.DocumentSymbol]]:
        """Called when a document symbols request it received."""
        ...



import re
from typing import Optional
from typing import Type

from docutils.parsers.rst import Directive
from lsprotocol.types import CompletionItem
from lsprotocol.types import CompletionItemKind
from lsprotocol.types import InsertTextFormat
from lsprotocol.types import Position
from lsprotocol.types import Range
from lsprotocol.types import TextEdit

from esbonio.lsp import CompletionContext

__all__ = ["render_directive_completion", "render_directive_option_completion"]


WORD = re.compile("[a-zA-Z]+")


def render_directive_completion(
    context: CompletionContext,
    name: str,
    directive: Type[Directive],
) -> Optional[CompletionItem]:
    """Render the given directive as a ``CompletionItem`` according to the current
    context.

    Parameters
    ----------
    context
       The context in which the completion should be rendered.

    name
       The name of the directive, as it appears in an rst file.

    directive
       The class that implements the directive.

    Returns
    -------
    Optional[CompletionItem]
       The final completion item or ``None``.
       If ``None`` is returned, then the given completion should be skipped.
    """

    if context.config.preferred_insert_behavior == "insert":
        return _render_directive_with_insert_text(context, name, directive)

    return _render_directive_with_text_edit(context, name, directive)


def render_directive_option_completion(
    context: CompletionContext,
    name: str,
    directive: str,
    implementation: Type[Directive],
) -> Optional[CompletionItem]:
    """Render the given directive option as a ``CompletionItem`` according to the
    current context.

    Parameters
    ----------
    context
       The context in which the completion should be rendered.

    name
       The name of the option, as it appears in an rst file.

    directive
       The name of the directive, as it appears in an rst file.

    implementation
       The class implementing the directive.

    Returns
    -------
    Optional[CompletionItem]
       The final completion item or ``None``.
       If ``None`` is returned, the given completion should be skipped.
    """

    if context.config.preferred_insert_behavior == "insert":
        return _render_directive_option_with_insert_text(
            context, name, directive, implementation
        )

    return _render_directive_option_with_text_edit(
        context, name, directive, implementation
    )


def _render_directive_with_insert_text(
    context: CompletionContext,
    name: str,
    directive: Type[Directive],
) -> Optional[CompletionItem]:
    """Render a ``CompletionItem`` using ``insertText`` fields.

    This implements the ``insert`` behavior for directives.
    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the directive, as it appears in an rst file.

    directive
       The class implementing the directive.

    """
    insert_text = f".. {name}::"
    user_text = context.match.group(0).strip()

    # Since we can't replace any existing text, it only makes sense
    # to offer completions that ailgn with what the user has already written.
    if not insert_text.startswith(user_text):
        return None

    # Except that's not entirely true... to quote the LSP spec. (emphasis added)
    #
    # > in the model the client should filter against what the user has already typed
    # > **using the word boundary rules of the language** (e.g. resolving the word
    # > under the cursor position). The reason for this mode is that it makes it
    # > extremely easy for a server to implement a basic completion list and get it
    # > filtered on the client.
    #
    # So in other words... if the cursor is inside a word, that entire word will be
    # replaced with what we have in `insert_text` so we need to be able to do something
    # like
    #    ..         -> image::
    #    .. im      -> image::
    #
    #    ..         -> code-block::
    #    .. cod     -> code-block::
    #    .. code-bl -> block::
    #
    #    ..         -> c:function::
    #    .. c       -> c:function::
    #    .. c:      -> function::
    #    .. c:fun   -> function::
    #
    # And since the client is free to interpret this how it likes, it's unlikely we'll
    # be able to get this right in all cases for all clients. So for now this is going
    # to target Kate's interpretation since it currently does not support ``text_edit``
    # and it was the editor that prompted this to be implemented in the first place.
    #
    # See: https://github.com/swyddfa/esbonio/issues/471

    # If the existing text ends with a delimiter, then we should simply remove the
    # entire prefix
    if user_text.endswith((":", "-", " ")):
        start_index = len(user_text)

    # Look for groups of word chars, replace text until the start of the final group
    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_directive_common(name, directive)
    item.insert_text = insert_text[start_index:]
    return item


def _render_directive_with_text_edit(
    context: CompletionContext,
    name: str,
    directive: Type[Directive],
) -> Optional[CompletionItem]:
    """Render a directive's ``CompletionItem`` using the ``textEdit`` field.

    This implements the ``replace`` insert behavior for directives.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the directive, as it appears in an rst file.

    directive
       The class implementing the directive.

    """
    match = context.match

    # Calculate the range of text the CompletionItems should edit.
    # If there is an existing argument to the directive, we should leave it untouched
    # otherwise, edit the whole line to insert any required arguments.
    start = match.span()[0] + match.group(0).find(".")
    include_argument = context.snippet_support
    end = match.span()[1]

    if match.group("argument"):
        include_argument = False
        end = match.span()[0] + match.group(0).find("::") + 2

    range_ = Range(
        start=Position(line=context.position.line, character=start),
        end=Position(line=context.position.line, character=end),
    )

    # TODO: Give better names to arguments based on what they represent.
    if include_argument:
        insert_format = InsertTextFormat.Snippet
        nargs = getattr(directive, "required_arguments", 0)
        args = " " + " ".join("${{{0}:arg{0}}}".format(i) for i in range(1, nargs + 1))
    else:
        args = ""
        insert_format = InsertTextFormat.PlainText

    insert_text = f".. {name}::{args}"

    item = _render_directive_common(name, directive)
    item.filter_text = insert_text
    item.text_edit = TextEdit(range=range_, new_text=insert_text)
    item.insert_text_format = insert_format

    return item


def _render_directive_common(
    name: str,
    directive: Type[Directive],
) -> CompletionItem:
    """Render the common fields of a directive's completion item."""

    try:
        dotted_name = f"{directive.__module__}.{directive.__name__}"
    except AttributeError:
        dotted_name = f"{directive.__module__}.{directive.__class__.__name__}"

    return CompletionItem(
        label=name,
        detail=dotted_name,
        kind=CompletionItemKind.Class,
        data={"completion_type": "directive"},
    )


def _render_directive_option_with_insert_text(
    context: CompletionContext,
    name: str,
    directive: str,
    implementation: Type[Directive],
) -> Optional[CompletionItem]:
    """Render a directive option's ``CompletionItem`` using the ``insertText`` field.

    This implements the ``insert`` insert behavior for directive options.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the directive option, as it appears in an rst file.

    directive
       The name of the directive, as it appears in an rst file.

    implementation
       The class implementing the directive.

    """

    insert_text = f":{name}:"
    user_text = context.match.group(0).strip()

    if not insert_text.startswith(user_text):
        return None

    if user_text.endswith((":", "-", " ")):
        start_index = len(user_text)

    else:
        start_indices = [m.start() for m in WORD.finditer(user_text)] or [
            len(user_text)
        ]
        start_index = max(start_indices)

    item = _render_directive_option_common(name, directive, implementation)
    item.insert_text = insert_text[start_index:]
    return item


def _render_directive_option_with_text_edit(
    context: CompletionContext,
    name: str,
    directive: str,
    implementation: Type[Directive],
) -> CompletionItem:
    """Render a directive option's ``CompletionItem`` using the``textEdit`` field.

    This implements the ``replace`` insert behavior for directive options.

    Parameters
    ----------
    context
       The context in which the completion is being generated.

    name
       The name of the directive option, as it appears in an rst file.

    directive
       The name of the directive, as it appears in an rst file.

    implementation
       The class implementing the directive.

    """

    match = context.match
    groups = match.groupdict()

    option = groups["option"]
    start = match.span()[0] + match.group(0).find(option)
    end = start + len(option)

    range_ = Range(
        start=Position(line=context.position.line, character=start),
        end=Position(line=context.position.line, character=end),
    )

    insert_text = f":{name}:"

    item = _render_directive_option_common(name, directive, implementation)
    item.filter_text = insert_text
    item.text_edit = TextEdit(range=range_, new_text=insert_text)

    return item


def _render_directive_option_common(
    name: str, directive: str, impl: Type[Directive]
) -> CompletionItem:
    """Render the common fields of a directive option's completion item."""

    try:
        impl_name = f"{impl.__module__}.{impl.__name__}"
    except AttributeError:
        impl_name = f"{impl.__module__}.{impl.__class__.__name__}"

    return CompletionItem(
        label=name,
        detail=f"{impl_name}:{name}",
        kind=CompletionItemKind.Field,
        data={"completion_type": "directive_option", "for_directive": directive},
    )



"""Extra support for roles added by sphinx."""
import json
import os.path
from typing import List
from typing import Optional

import pygls.uris as Uri
from lsprotocol.types import CompletionItem

from esbonio.lsp.roles import Roles
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util import resources
from esbonio.lsp.util.filepaths import complete_sphinx_filepaths
from esbonio.lsp.util.filepaths import path_to_completion_item


class Downloads:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_targets(
        self, context: CompletionContext, name: str, domain: Optional[str]
    ) -> List[CompletionItem]:
        if domain or name != "download":
            return []

        if not self.rst.app:
            return []

        srcdir = self.rst.app.srcdir
        partial = context.match.group("label")
        base = os.path.dirname(Uri.to_fs_path(context.doc.uri))
        items = complete_sphinx_filepaths(srcdir, base, partial)  # type: ignore[arg-type]

        return [path_to_completion_item(context, p) for p in items]


def esbonio_setup(rst: SphinxLanguageServer, roles: Roles):
    sphinx_docs = resources.read_string("esbonio.lsp.sphinx", "roles.json")
    roles.add_documentation(json.loads(sphinx_docs))

    roles.add_target_completion_provider(Downloads(rst))



import json
from typing import Any
from typing import Dict
from typing import Optional

import docutils.parsers.rst.roles as docutils_roles

from esbonio.lsp.roles import RoleLanguageFeature
from esbonio.lsp.roles import Roles
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.util import resources


class Docutils(RoleLanguageFeature):
    """Support for docutils' built-in roles."""

    def __init__(self) -> None:
        self._roles: Optional[Dict[str, Any]] = None
        """Cache for known roles."""

    @property
    def roles(self) -> Dict[str, Any]:
        if self._roles is not None:
            return self._roles

        found_roles = {**docutils_roles._roles, **docutils_roles._role_registry}

        self._roles = {
            k: v
            for k, v in found_roles.items()
            if v != docutils_roles.unimplemented_role
        }

        return self._roles

    def get_implementation(self, role: str, domain: str):
        if domain:
            return None

        return self.roles.get(role, None)

    def index_roles(self) -> Dict[str, Any]:
        return self.roles


def esbonio_setup(rst: RstLanguageServer, roles: Roles):
    documentation = resources.read_string("esbonio.lsp.rst", "roles.json")

    roles.add_documentation(json.loads(documentation))
    roles.add_feature(Docutils())



import argparse
import logging
import sys
import warnings
from typing import Union

from lsprotocol import types
from pygls.protocol import default_converter

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


def esbonio_converter():
    converter = default_converter()
    converter.register_structure_hook(Union[Literal["auto"], int], lambda obj, _: obj)

    def position_hook(obj, type_):
        """Parse a position, gracefully handling invalid data"""
        l: int = obj.get("line", None) or 0
        c: int = obj.get("character", None) or 0

        # Clamp positions to valid range.
        return types.Position(
            line=min(max(l, 0), 2147483647),
            character=min(max(c, 0), 2147483647),
        )

    converter.register_structure_hook(types.Position, position_hook)
    return converter


def setup_cli(progname: str, description: str) -> argparse.ArgumentParser:
    """Return an argument parser with the default command line options required for
    main.
    """

    cli = argparse.ArgumentParser(prog=f"python -m {progname}", description=description)
    cli.add_argument(
        "-p",
        "--port",
        type=int,
        default=None,
        help="start a TCP instance of the language server listening on the given port.",
    )
    cli.add_argument(
        "--version", action="store_true", help="print the current version and exit."
    )

    modules = cli.add_argument_group(
        "modules", "include/exclude language server modules."
    )
    modules.add_argument(
        "-i",
        "--include",
        metavar="MOD",
        action="append",
        default=[],
        dest="included_modules",
        help="include an additional module in the server configuration, can be given multiple times.",
    )
    modules.add_argument(
        "-e",
        "--exclude",
        metavar="MOD",
        action="append",
        default=[],
        dest="excluded_modules",
        help="exclude a module from the server configuration, can be given multiple times.",
    )

    return cli


def main(cli: argparse.ArgumentParser):
    """Standard main function for each of the default language servers."""

    # Put these here to avoid circular import issues.
    from esbonio.lsp import __version__
    from esbonio.lsp import create_language_server
    from esbonio.lsp.log import LOG_NAMESPACE
    from esbonio.lsp.log import MemoryHandler

    args = cli.parse_args()

    if args.version:
        print(f"v{__version__}")
        sys.exit(0)

    modules = list(args.modules)

    for mod in args.included_modules:
        modules.append(mod)

    for mod in args.excluded_modules:
        if mod in modules:
            modules.remove(mod)

    # Ensure we can capture warnings.
    logging.captureWarnings(True)
    warnlog = logging.getLogger("py.warnings")

    if not sys.warnoptions:
        warnings.simplefilter("default")  # Enable capture of DeprecationWarnings

    # Setup a temporary logging handler that can cache messages until the language server
    # is ready to forward them onto the client.
    logger = logging.getLogger(LOG_NAMESPACE)
    logger.setLevel(logging.DEBUG)

    handler = MemoryHandler()
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    warnlog.addHandler(handler)

    server = create_language_server(
        args.server_cls,
        modules,
        name="esbonio",
        version=__version__,
        # TODO: Figure out how to make this extensible
        converter_factory=esbonio_converter,
    )

    if args.port:
        server.start_tcp("localhost", args.port)
    else:
        server.start_io()



import re

DIRECTIVE: "re.Pattern" = re.compile(
    r"""
    (\s*)                             # directives can be indented
    (?P<directive>
      \.\.                            # directives start with a comment
      [ ]?                            # followed by a space
      (?P<substitution>\|             # this could be a substitution definition
        (?P<substitution_text>[^|]+)?
      \|?)?
      [ ]?
      ((?P<domain>[\w]+):(?!:))?      # directives may include a domain
      (?P<name>([\w-]|:(?!:))+)?      # directives have a name
      (::)?                           # directives end with '::'
    )
    ([\s]+(?P<argument>.*?)\s*$)?     # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete directives.

This does **not** include any options or content that may be included underneath
the initial declaration. A number of named capture groups are available.

``name``
   The name of the directive, not including the domain prefix.

``domain``
   The domain prefix

``directive``
   Everything that makes up a directive, from the initial ``..`` up to and including the
   ``::`` characters.

``argument``
   All argument text.

``substitution``
   If the directive is part of a substitution definition, this group will contain

**Example**

Here is an example with a "standard" directive

.. include:: ../../../lib/esbonio/tests/doctests/example_directive_pattern.txt

And here is an example with a substitution definition

.. include:: ../../../lib/esbonio/tests/doctests/example_directive_substitution_pattern.txt

"""


DIRECTIVE_OPTION: "re.Pattern" = re.compile(
    r"""
    (?P<indent>\s+)       # directive options must be indented
    (?P<option>
      :                   # options start with a ':'
      (?P<name>[\w-]+)?   # options have a name
      :?                  # options end with a ':'
    )
    (\s*
      (?P<value>.*)       # options can have a value
    )?
    """,
    re.VERBOSE,
)
"""A regular expression used to detect and parse partial and complete directive options.

A number of named capture groups are available

``name``
   The name of the option

``option``
   The name of the option including the surrounding ``:`` characters.

``indent``
   The whitespace characters making preceeding the initial ``:`` character

``value``
   The value passed to the option

**Example**

.. include:: ../../../lib/esbonio/tests/doctests/example_directive_option_pattern.txt

"""


ROLE = re.compile(
    r"""
    ([^\w:]|^\s*)                     # roles cannot be preceeded by letter chars
    (?P<role>
      :                               # roles begin with a ':' character
      (?!:)                           # the next character cannot be a ':'
      ((?P<domain>[\w]+):(?=\w))?     # roles may include a domain (that must be followed by a word character)
      ((?P<name>[\w-]+):?)?           # roles have a name
    )
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )?
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete roles.

I'm not sure if there are offical names for the components of a role, but the
language server breaks a role down into a number of parts::

                 vvvvvv label
                v modifier(optional)
               vvvvvvvv target
   :c:function:`!malloc`
   ^^^^^^^^^^^^ role
      ^^^^^^^^ name
    ^ domain (optional)

The language server sometimes refers to the above as a "plain" role, in that the
role's target contains just the label of the object it is linking to. However it's
also possible to define "aliased" roles, where the link text in the final document
is overriden, for example::

                vvvvvvvvvvvvvvvvvvvvvvvv alias
                                          vvvvvv label
                                         v modifier (optional)
               vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv target
   :c:function:`used to allocate memory <~malloc>`
   ^^^^^^^^^^^^ role
      ^^^^^^^^ name
    ^ domain (optional)

See :func:`tests.test_roles.test_role_regex` for a list of example strings this pattern
is expected to match.
"""


DEFAULT_ROLE = re.compile(
    r"""
    (?<![:`])
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete "default" roles.

A "default" role is the target part of a normal role - but without the ``:name:`` part.
"""



import importlib
import json
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives as docutils_directives

from esbonio.lsp.directives import DirectiveLanguageFeature
from esbonio.lsp.directives import Directives
from esbonio.lsp.util import resources


class Docutils(DirectiveLanguageFeature):
    """Support for docutils' built-in directives."""

    def __init__(self) -> None:
        self._directives: Optional[Dict[str, Type[Directive]]] = None
        """Cache for known directives."""

    @property
    def directives(self) -> Dict[str, Type[Directive]]:
        if self._directives is not None:
            return self._directives

        ignored_directives = ["restructuredtext-test-directive"]
        found_directives = {
            **docutils_directives._directive_registry,
            **docutils_directives._directives,
        }

        self._directives = {
            k: resolve_directive(v)
            for k, v in found_directives.items()
            if k not in ignored_directives
        }

        return self._directives

    def get_implementation(self, directive: str, domain: Optional[str]):
        if domain:
            return None

        return self.directives.get(directive, None)

    def index_directives(self) -> Dict[str, Type[Directive]]:
        return self.directives


def resolve_directive(
    directive: Union[Type[Directive], Tuple[str, str]]
) -> Type[Directive]:
    """Return the directive based on the given reference.

    'Core' docutils directives are returned as tuples ``(modulename, ClassName)``
    so they need to be resolved manually.
    """

    if isinstance(directive, tuple):
        mod, cls = directive

        modulename = f"docutils.parsers.rst.directives.{mod}"
        module = importlib.import_module(modulename)
        return getattr(module, cls)

    return directive


def esbonio_setup(directives: Directives):
    documentation = resources.read_string("esbonio.lsp.rst", "directives.json")

    directives.add_documentation(json.loads(documentation))
    directives.add_feature(Docutils())



import inspect
import logging
import os
import pathlib
import sys
from types import ModuleType
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

from sphinx.util.logging import OnceFilter
from sphinx.util.logging import SphinxLogRecord
from sphinx.util.logging import WarningLogRecordTranslator

from . import types
from .util import logger
from .util import send_message

DIAGNOSTIC_SEVERITY = {
    logging.ERROR: types.DiagnosticSeverity.Error,
    logging.INFO: types.DiagnosticSeverity.Information,
    logging.WARNING: types.DiagnosticSeverity.Warning,
}


class SphinxLogHandler(logging.Handler):
    """A logging handler that can extract errors from Sphinx's build output."""

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app = app
        self.translator = WarningLogRecordTranslator(app)
        self.only_once = OnceFilter()
        self.diagnostics: Dict[str, Set[types.Diagnostic]] = {}

    def get_location(self, location: str) -> Tuple[str, Optional[int]]:
        if not location:
            conf = pathlib.Path(self.app.confdir, "conf.py")
            return (str(conf), None)

        lineno = None
        path, parts = self.get_location_path(location)

        if len(parts) == 1:
            try:
                lineno = int(parts[0])
            except ValueError:
                pass

        if len(parts) == 2 and parts[0].startswith("docstring of "):
            target = parts[0].replace("docstring of ", "")
            lineno = self.get_docstring_location(target, parts[1])

        return (path, lineno)

    def get_location_path(self, location: str) -> Tuple[str, List[str]]:
        """Determine the filepath from the given location."""

        if location.startswith("internal padding before "):
            location = location.replace("internal padding before ", "")

        if location.startswith("internal padding after "):
            location = location.replace("internal padding after ", "")

        path, *parts = location.split(":")

        # On windows the rest of the path will be the first element of parts
        if pathlib.Path(location).drive:
            path += f":{parts.pop(0)}"

        # Diagnostics in .. included:: files are reported relative to the process'
        # working directory, so ensure the path is absolute.
        path = os.path.abspath(path)

        return path, parts

    def get_docstring_location(self, target: str, offset: str) -> Optional[int]:
        # The containing module will be the longest substring we can find in target
        candidates = [m for m in sys.modules.keys() if target.startswith(m)] + [""]
        module = sys.modules.get(sorted(candidates, key=len, reverse=True)[0], None)

        if module is None:
            return None

        obj: Union[ModuleType, Any, None] = module
        dotted_name = target.replace(module.__name__ + ".", "")

        for name in dotted_name.split("."):
            obj = getattr(obj, name, None)
            if obj is None:
                return None

        try:
            _, line = inspect.getsourcelines(obj)  # type: ignore

            # Correct off by one error for docstrings that don't start with a newline.
            nl = (obj.__doc__ or "").startswith("\n")
            return line + int(offset) - (not nl)
        except Exception:
            logger.debug("Unable to determine diagnostic location\n%s", exc_info=True)
            return None

    def emit(self, record: logging.LogRecord) -> None:
        conditions = [
            "sphinx" not in record.name,
            record.levelno not in {logging.WARNING, logging.ERROR},
        ]

        if any(conditions):
            # Log the record as normal
            self.do_emit(record)
            return

        # Let sphinx extract location info for warning/error messages
        self.translator.filter(record)  # type: ignore

        # Only process errors/warnings once.
        # Note: This isn't a silver bullet as it only catches messages that are explicitly
        #       marked as to be logged only once e.g. logger.warning(..., once=True).
        if not self.only_once.filter(record):
            return

        loc = record.location if isinstance(record, SphinxLogRecord) else ""
        doc, lineno = self.get_location(loc)
        line = lineno or 1

        try:
            # Not every message contains a string...
            if not isinstance(record.msg, str):
                message = str(record.msg)
            else:
                message = record.msg

            # Only attempt to format args if there are args to format
            if record.args is not None and len(record.args) > 0:
                message = message % record.args

        except Exception:
            message = str(record.msg)
            logger.error("Unable to format diagnostic message: %s", exc_info=True)

        diagnostic = types.Diagnostic(
            range=types.Range(
                start=types.Position(line=line - 1, character=0),
                end=types.Position(line=line, character=0),
            ),
            message=message,
            severity=DIAGNOSTIC_SEVERITY.get(
                record.levelno, types.DiagnosticSeverity.Warning
            ),
        )

        self.diagnostics.setdefault(doc, set()).add(diagnostic)
        self.do_emit(record)

    def do_emit(self, record):
        params = types.LogMessageParams(message=self.format(record).strip(), type=4)
        send_message(types.LogMessage(params=params))
